from alphastats.loader.MaxQuantLoader import MaxQuantLoader
from alphastats.loader.AlphaPeptLoader import AlphaPeptLoader
from alphastats.loader.FragPipeLoader import FragPipeLoader
from alphastats.loader.DIANNLoader import DIANNLoader
from alphastats.loader.SpectronautLoader import SpectronautLoader

software_options = {
    "MaxQuant": {
        "import_file": "proteinGroups.txt",
        "intensity_column": ["LFQ intensity [sample]"],
        "index_column": ["Protein IDs"],
        "loader_function": MaxQuantLoader,
    },
    "AlphaPept": {
        "import_file": "results_proteins.csv or results.hdf",
        "intensity_column": ["[sample]_LFQ"],
        "index_column": ["Unnamed: 0"],
        "loader_function": AlphaPeptLoader,
    },
    "DIANN": {
        "import_file": "report.pg_matrix.tsv",
        "intensity_column": ["[sample]"],
        "index_column": ["Protein.Group"],
        "loader_function": DIANNLoader,
    },
    "FragPipe": {
        "import_file": "combined_protein.tsv",
        "intensity_column": ["[sample] MaxLFQ Intensity", "[sample] Intensity"],
        "index_column": ["Protein", "Protein ID"],
        "loader_function": FragPipeLoader,
    },
    "Spectronaut": {
        "import_file": "spectronaut.tsv",
        "intensity_column": ["PG.Quantity", "F.PeakArea", "PG.Log2Quantity"],
        "index_column": ["PG.ProteinGroups", "PEP.StrippedSequence"],
        "loader_function": SpectronautLoader,
    },
}



from codecs import ignore_errors
from itertools import permutations
import pandas as pd
import scipy.stats
import numpy as np
import pingouin
from alphastats.utils import ignore_warning
from tqdm import tqdm
from functools import lru_cache
from typing import Union

from alphastats.statistics.DifferentialExpressionAnalysis import DifferentialExpressionAnalysis
from alphastats.statistics.Anova import Anova


class Statistics:
    def _calculate_foldchange(self, mat_transpose:pd.DataFrame, group1_samples:list, group2_samples:list) -> pd.DataFrame:
        mat_transpose += 0.00001

        if self.preprocessing_info["Log2-transformed"]:
            fc = (
                mat_transpose[group1_samples].T.mean().values
                - mat_transpose[group2_samples].T.mean().values
            )
        
        else:
            fc = (
                mat_transpose[group1_samples].T.mean().values
                / mat_transpose[group2_samples].T.mean().values
            )
            fc = np.log2(fc)

        return pd.DataFrame({"log2fc": fc, self.index_column: mat_transpose.index})
    
    def _add_metadata_column(self, group1_list: list, group2_list: list):

        # create new column in metadata with defined groups
        metadata = self.metadata

        sample_names = metadata[self.sample].to_list()
        misc_samples = list(set(group1_list + group2_list) - set(sample_names))
        if len(misc_samples) > 0:
            raise ValueError(
                f"Sample names: {misc_samples} are not described in Metadata."
            )

        column = "_comparison_column"
        conditons = [
            metadata[self.sample].isin(group1_list),
            metadata[self.sample].isin(group2_list),
        ]
        choices = ["group1", "group2"]
        metadata[column] = np.select(conditons, choices, default=np.nan)
        self.metadata = metadata

        return column, "group1", "group2"
    
    @ignore_warning(RuntimeWarning)
    def diff_expression_analysis(
            self,
            group1: Union[str, list],
            group2: Union[str, list],
            column: str = None,
            method: str = "ttest",
            perm: int = 10,
            fdr: float = 0.05,
        ) -> pd.DataFrame:
        """Perform differential expression analysis doing a a t-test or Wald test. A wald test will fit a generalized linear model.

        Args:
            column (str): column name in the metadata file with the two groups to compare
            group1 (str/list): name of group to compare needs to be present in column or list of sample names to compare
            group2 (str/list): name of group to compare needs to be present in column  or list of sample names to compare
            method (str,optional): statistical method to calculate differential expression, for Wald-test 'wald', paired t-test 'paired-ttest'. Default 'ttest'

        Returns:
            pandas.DataFrame:
            pandas Dataframe with foldchange, foldchange_log2 and pvalue
            for each ProteinID/ProteinGroup between group1 and group2.

            * ``'Protein ID'``: ProteinID/ProteinGroup
            * ``'pval'``: p-value of the ProteinID/ProteinGroup
            * ``'qval'``: multiple testing - corrected p-value
            * ``'log2fc'``: log2(foldchange)
            * ``'grad'``: the gradient of the log-likelihood
            * ``'coef_mle'``: the maximum-likelihood estimate of coefficient in liker-space
            * ``'coef_sd'``: the standard deviation of the coefficient in liker-space
            * ``'ll'``: the log-likelihood of the estimation
        """

        df = DifferentialExpressionAnalysis(
            dataset=self, 
            group1=group1, 
            group2=group2, 
            column=column, method=method,
            perm=perm, 
            fdr=fdr
        ).perform()
        return df

    @ignore_warning(RuntimeWarning)
    def tukey_test(self, protein_id:str, group:str, df: pd.DataFrame=None) -> pd.DataFrame:
        """Calculate Pairwise Tukey-HSD post-hoc test
        Wrapper around:
        https://pingouin-stats.org/generated/pingouin.pairwise_tukey.html#pingouin.pairwise_tukey

        Args:
            protein_id (str): ProteinID to calculate Pairwise Tukey-HSD post-hoc test - dependend variable
            group (str): A metadata column used calculate pairwise tukey
            df (pandas.DataFrame, optional): Defaults to None.

        Returns:
            pandas.DataFrame:
            * ``'A'``: Name of first measurement
            * ``'B'``: Name of second measurement
            * ``'mean(A)'``: Mean of first measurement
            * ``'mean(B)'``: Mean of second measurement
            * ``'diff'``: Mean difference (= mean(A) - mean(B))
            * ``'se'``: Standard error
            * ``'T'``: T-values
            * ``'p-tukey'``: Tukey-HSD corrected p-values
            * ``'hedges'``: Hedges effect size (or any effect size defined in
            ``effsize``)
            * ``'comparison'``: combination of measurment
            * ``'Protein ID'``: ProteinID/ProteinGroup
        """
        if df is None:
            df = (
                self.mat[[protein_id]]
                .reset_index()
                .rename(columns={"index": self.sample})
            )
            df = df.merge(self.metadata, how="inner", on=[self.sample])

        try:
            tukey_df = pingouin.pairwise_tukey(data=df, dv=protein_id, between=group)
            tukey_df["comparison"] = (
                tukey_df["A"] + " vs. " + tukey_df["B"] + " Tukey Test"
            )
            tukey_df[self.index_column] = protein_id

        except ValueError:
            tukey_df = pd.DataFrame()

        return tukey_df

    @ignore_warning(RuntimeWarning)
    def anova(self, column:str, protein_ids="all", tukey: bool=True) -> pd.DataFrame:
        """One-way Analysis of Variance (ANOVA)

        Args:
            column (str): A metadata column used to calculate ANOVA
            protein_ids (str or list, optional): ProteinIDs to calculate ANOVA for - dependend variable either ProteinID as string, several ProteinIDs as list or "all" to calculate ANOVA for all ProteinIDs. Defaults to "all".
            tukey (bool, optional): Whether to calculate a Tukey-HSD post-hoc test. Defaults to True.

        Returns:
            pandas.DataFrame:
            * ``'Protein ID'``: ProteinID/ProteinGroup
            * ``'ANOVA_pvalue'``: p-value of ANOVA
            * ``'A vs. B Tukey test'``: Tukey-HSD corrected p-values (each combination represents a column)
        """
        return Anova(dataset=self, column=column, protein_ids=protein_ids, tukey=tukey).perform()
        

    @lru_cache(maxsize=20)
    def ancova(self, protein_id:str, covar: Union[str, list], between:str) -> pd.DataFrame:
        """Analysis of covariance (ANCOVA) with on or more covariate(s).
        Wrapper around = https://pingouin-stats.org/generated/pingouin.ancova.html

        Args:
            protein_id (str): ProteinID/ProteinGroup - dependent variable
            covar (str or list):   Name(s) of column(s) in metadata with the covariate.
            between (str): Name of column in data with the between factor.

        Returns:
            pandas.Dataframe:

            ANCOVA summary:

            * ``'Source'``: Names of the factor considered
            * ``'SS'``: Sums of squares
            * ``'DF'``: Degrees of freedom
            * ``'F'``: F-values
            * ``'p-unc'``: Uncorrected p-values
            * ``'np2'``: Partial eta-squared
        """
        df = self.mat[protein_id].reset_index().rename(columns={"index": self.sample})
        df = self.metadata.merge(df, how="inner", on=[self.sample])
        ancova_df = pingouin.ancova(df, dv=protein_id, covar=covar, between=between)
        return ancova_df



from alphastats.loader.BaseLoader import BaseLoader
import pandas as pd


class DIANNLoader(BaseLoader):
    """Loader for DIA-NN output files
    https://github.com/vdemichev/DiaNN
    """

    def __init__(
        self,
        file,
        intensity_column="[sample]",
        index_column="Protein.Group",
        sep="\t",
        **kwargs
    ):
        """Import DIA-NN output data report.pg_matrix.tsv

        Args:
            file (str): DIA-NN output file report.pg_matrix.tsv
            intensity_column (str, optional): columns containing the intensity column for each experiment. Defaults to "[experiment]".
            index_column (str, optional): column with the Protein IDs. Defaults to "Protein.Group".
            sep (str, optional): file separation of the input file. Defaults to "\t".
        """

        super().__init__(file, intensity_column, index_column, sep)
        self.software = "DIANN"
        self.no_sample_column = [
            "PG.Q.value",
            "Global.PG.Q.value",
            "PTM.Q.value",
            "PTM.Site.Confidence",
            "PG.Quantity",
            "Protein.Group",
            "Protein.Ids",
            "Protein.Names",
            "Genes",
            "First.Protein.Description",
            "contamination_library",
        ]
        self._remove_filepath_from_name()
        self._add_tag_to_sample_columns()
        self._add_contamination_column()
        self._read_all_columns_as_string()

    def _add_tag_to_sample_columns(self):
        """
        when creating matrix sample columns wont be found when it is only specified as [experiment]
        so tag will be added
        """
        # TODO this is very fragile as changes in column names can break this

        self.rawinput.columns = [
            str(col) + "_Intensity" if col not in self.no_sample_column else str(col)
            for col in self.rawinput.columns
        ]
        self.intensity_column = "[sample]_Intensity"

    @staticmethod
    def _split_path(file_path):
        """
        split file path for windows and macOS
        """
        # try:
        if "/" in file_path:
            file = file_path.split("/")[-1]

        else:
            file = file_path.split("\\")[-1]

        # windows path can cause error
        # except SyntaxError:
        #     file = file_path

        return file

    def _remove_filepath_from_name(self):
        """
        split filepath so only filename is used for analysis
        """

        self.rawinput.columns = [
            self._split_path(col) if col not in self.no_sample_column else str(col)
            for col in self.rawinput.columns
        ]



from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px

import math

class SampleHistogram:
    """
    Plot denisty plot of each sample in a matrix
    """
    def __init__(self, dataset):
        self.dataset = dataset
        self._get_matrix()
    
    def _get_matrix(self):
        self.data = self.dataset.mat.transpose()
    
    def _get_position_in_matrix(self, count):
        count += 1
        row = math.ceil(count / 4 )
        col = count % 4
            
        if col == 0:
            col = 4
        
        return col, row
    
    def _initalize_plot(self):
        n_samples = len(self.data.columns.to_list())
        
        if n_samples < 4:
            n_rows = 1
            n_columns = n_samples
        
        else:
            n_rows = math.ceil(n_samples/4)
            n_columns = 4
        
        self.plot = make_subplots(
            rows=n_rows, 
            cols=n_columns,
            subplot_titles=self.data.columns.to_list()
        )
    
    def plot(self):
        self._initalize_plot()
        
        for count, x in enumerate(self.data.columns):
            col, row = self._get_position_in_matrix(count)

            data_column = self.data[self.data[x] != 0][x].dropna()
           
            self.plot.add_trace(go.Histogram(x=data_column),row=row,col=col)
        
        self.plot.update_layout(showlegend=False)

        return self.plot




from alphastats.loader.BaseLoader import BaseLoader
import pandas as pd

# Philosopher
# class name needs to be discussed whether MSFragger/Fragpipe/Philospher
class FragPipeLoader(BaseLoader):
    """Loader for FragPipe-Philosopheroutputfiles
    https://fragpipe.nesvilab.org/docs/tutorial_fragpipe_outputs.html#combined_proteintsv
    """

    def __init__(
        self,
        file,
        intensity_column="[sample] MaxLFQ Intensity ",
        index_column="Protein",
        gene_names_column="Gene Names",
        confidence_column="Protein Probability",
        sep="\t",
        **kwargs
    ):

        super().__init__(file, intensity_column, index_column, sep)

        if gene_names_column in self.rawinput.columns.to_list():
            self.gene_names = gene_names_column

        self.confidence_column = confidence_column
        self.software = "MSFragger_Philosopher"


"""
- find way to deal with Indistinguishable Proteins, combine ProteinIDs?
"""
# https://github.com/Nesvilab/MSFragger/wiki/Interpreting-MSFragger-Output

#  SAMPLE
# columm Spectrum

# https://fragpipe.nesvilab.org/docs/tutorial_fragpipe_outputs.html#combined_proteintsv
# Frag pipe
# https://github.com/Nesvilab/philosopher/wiki/Combined-protein-reports
#  ProteinProphet: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5863791/



import streamlit as st


plotting_options = {
    "Sampledistribution Plot": {
        "settings": {
            "method": {"options": ["violin", "box"], "label": "Plot layout"},
            "color": {
                "options": [None] + st.session_state.metadata_columns,
                "label": "Color according to",
            },
        },
        "function": st.session_state.dataset.plot_sampledistribution,
    },
    "Intensity Plot": {
        "settings": {
            "protein_id": {
                "options": st.session_state.dataset.mat.columns.to_list(),
                "label": "ProteinID/ProteinGroup",
            },
            "method": {
                "options": ["violin", "box", "scatter"],
                "label": "Plot layout",
            },
            "group": {
                "options": [None] + st.session_state.metadata_columns,
                "label": "Color according to",
            },
        },
        "function": st.session_state.dataset.plot_intensity,
    },
    "PCA Plot": {
        "settings": {
            "group": {
                "options": [None] + st.session_state.metadata_columns,
                "label": "Color according to",
            },
            "circle": {"label": "Circle"},
        },
        "function": st.session_state.dataset.plot_pca,
    },
    "UMAP Plot": {
        "settings": {
            "group": {
                "options": [None] + st.session_state.metadata_columns,
                "label": "Color according to",
            },
            "circle": {"label": "Circle"},
        },
        "function": st.session_state.dataset.plot_umap,
    },
    "t-SNE Plot": {
        "settings": {
            "group": {
                "options": [None] + st.session_state.metadata_columns,
                "label": "Color according to",
            },
            "circle": {"label": "Circle"},
        },
        "function": st.session_state.dataset.plot_tsne,
    },
    "Volcano Plot": {
        "between_two_groups": True,
        "function": st.session_state.dataset.plot_volcano,
    },
    "Clustermap": {"function": st.session_state.dataset.plot_clustermap},
    "Dendrogram": {"function": st.session_state.dataset.plot_dendrogram},
}

statistic_options = {
    "Differential Expression Analysis - T-test": {
        "between_two_groups": True,
        "function": st.session_state.dataset.diff_expression_analysis,
    },
    "Differential Expression Analysis - Wald-test": {
        "between_two_groups": True,
        "function": st.session_state.dataset.diff_expression_analysis,
    },
    "Tukey - Test": {
        "settings": {
            "protein_id": {
                "options": st.session_state.dataset.mat.columns.to_list(),
                "label": "ProteinID/ProteinGroup",
            },
            "group": {
                "options": st.session_state.metadata_columns,
                "label": "A metadata variable to calculate pairwise tukey",
            },
        },
        "function": st.session_state.dataset.tukey_test,
    },
    "ANOVA": {
        "settings": {
            "column": {
                "options": st.session_state.metadata_columns,
                "label": "A variable from the metadata to calculate ANOVA",
            },
            "protein_ids": {
                "options": ["all"] + st.session_state.dataset.mat.columns.to_list(),
                "label": "All ProteinIDs/or specific ProteinID to perform ANOVA",
            },
            "tukey": {"label": "Follow-up Tukey"},
        },
        "function": st.session_state.dataset.anova,
    },
    "ANCOVA": {
        "settings": {
            "protein_id": {
                "options": [None] + st.session_state.dataset.mat.columns.to_list(),
                "label": "Color according to",
            },
            "covar": {
                "options": st.session_state.metadata_columns,
                "label": "Name(s) of column(s) in metadata with the covariate.",
            },
            "between": {
                "options": st.session_state.metadata_columns,
                "label": "Name of the column in the metadata with the between factor.",
            },
        },
        "function": st.session_state.dataset.ancova,
    },
}



from alphastats.plots.PlotUtils import PlotUtils, seaborn_object

import seaborn as sns
import random


class ClusterMap(PlotUtils):
    def __init__(
        self,
        dataset,
        label_bar,
        only_significant,
        group,
        subgroups
        ):
        self.dataset = dataset
        self.label_bar = label_bar
        self.only_significant = only_significant
        self.group = group
        self.subgroups = subgroups

        self._prepare_df()
        self._plot()


    def _prepare_df(self):
        df = self.dataset.mat.loc[:, (self.dataset.mat != 0).any(axis=0)]

        if self.group is not None and self.subgroups is not None:
            metadata_df = self.dataset.metadata[
                self.dataset.metadata[self.group].isin(self.subgroups + [self.dataset.sample])
            ]
            samples = metadata_df[self.dataset.sample]
            df = df.filter(items=samples, axis=0)

        else:
            metadata_df = self.dataset.metadata

        if self.only_significant and self.group is not None:
            anova_df = self.dataset.anova(column=self.group, tukey=False)
            significant_proteins = anova_df[anova_df["ANOVA_pvalue"] < 0.05][
                self.dataset.index_column
            ].to_list()
            df = df[significant_proteins]

        if self.label_bar is not None:
            self._create_label_bar(
               metadata_df
            )

        self.prepared_df = self.dataset.mat.loc[:, (self.dataset.mat != 0).any(axis=0)].transpose()
  

    def _plot(self):
        fig = sns.clustermap(self.prepared_df, col_colors=self.label_bar)

        if self.label_bar is not None:
           fig = self._add_label_bar(fig)

        # set attributes
        setattr(fig, "plotting_data", self.prepared_df)
        setattr(fig, "preprocessing", self.dataset.preprocessing_info)
        setattr(fig, "method", "clustermap")

        self.plot = fig

    def _add_label_bar(self, fig):
        for label in self.s.unique():
                fig.ax_col_dendrogram.bar(
                    0, 0, color=self.lut[label], label=label, linewidth=0
                )
                fig.ax_col_dendrogram.legend(loc="center", ncol=6)
        return fig

    def _create_label_bar(self, metadata_df):
        colorway = [
            "#009599",
            "#005358",
            "#772173",
            "#B65EAF",
            "#A73A00",
            "#6490C1",
            "#FF894F",
        ]

        self.s = metadata_df[self.label_bar]
        su = self.s.unique()
        colors = sns.light_palette(random.choice(colorway), len(su))
        self.lut = dict(zip(su, colors))
        self.label_bar = self.s.map(self.lut)

       



from alphastats.loader.BaseLoader import BaseLoader
import pandas as pd
import numpy as np
import logging



class SpectronautLoader(BaseLoader):
    """Loader for Spectronaut outputfiles 
    """

    def __init__(
        self,
        file,
        intensity_column="PG.Quantity",
        index_column="PG.ProteinGroups", 
        sample_column = "R.FileName",
        gene_names_column="PG.Genes",
        filter_qvalue = True,
        qvalue_cutoff = 0.01,
        sep="\t"
    ):
        """Loads Spectronaut output. Will add contamination column for further analysis.

        Args:
            file (str): path to Spectronaut outputfile or pandas.DataFrame 
            intensity_column (str, optional): columns where the intensity of the proteins are given. Defaults to "PG.Quantity".
            index_column (str, optional): column indicating the protein groups. Defaults to "PG.ProteinGroups".
            sample_column (str, optional): column that contains sample names used for downstream analysis. Defaults to "R.FileName".
            gene_names_column (str, optional): column with gene names. Defaults to "PG.Genes".
            filter_qvalue (bool, optional): will filter out the intensities that have greater than qvalue_cutoff in EG.Qvalue column. Those intensities will be replaced with zero and will be considered as censored missing values for imputation purpose.. Defaults to True.
            qvalue_cutoff (float, optional): cut off vaéie. Defaults to 0.01.
            sep (str, optional): file separation of file. Defaults to "\t".
        """
        
        self.software = "Spectronaut"
        self.intensity_column = intensity_column
        self.index_column = index_column
        self.confidence_column = None
        self.filter_columns = []
        self.evidence_df = None 
        self.gene_names = None

        self._read_spectronaut_file(file=file, sep=sep)

        if filter_qvalue:
            self._filter_qvalue(qvalue_cutoff=qvalue_cutoff)

        self._reshape_spectronaut(sample_column=sample_column, gene_names_column=gene_names_column)
        self._add_contamination_column()
        self._read_all_columns_as_string()
              

    def _reshape_spectronaut(self, sample_column, gene_names_column):
        """
        other proteomics softwares use a wide format (column for each sample)
        reshape to a wider format
        """
        self.rawinput["sample"] = self.rawinput[sample_column] + "_" + self.intensity_column
        
        indexing_columns = [self.index_column]
        
        if gene_names_column in self.rawinput.columns.to_list():
            self.gene_names = gene_names_column
            indexing_columns += [self.gene_names]
        
        keep_columns = [self.intensity_column, "sample"] + indexing_columns
        
        df = self.rawinput[keep_columns].drop_duplicates()
        df = df.pivot(columns='sample', index=indexing_columns, values=self.intensity_column)
        df.reset_index(inplace=True)
        
        self.rawinput = df
        
        self.intensity_column = "[sample]_" + self.intensity_column

    def _filter_qvalue(self, qvalue_cutoff):
        if "EG.Qvalue" not in self.rawinput.columns.to_list():
            raise Warning("Column EG.Qvalue not found in file. File will not be filtered according to q-value.")
        
        rows_before_filtering = self.rawinput.shape[0]
        self.rawinput = self.rawinput[self.rawinput["EG.Qvalue"] < qvalue_cutoff]
        rows_after_filtering = self.rawinput.shape[0]

        rows_removed = rows_before_filtering - rows_after_filtering
        logging.info(f"{rows_removed} identification with a qvalue below {qvalue_cutoff} have been removed")
        
    
    def _read_spectronaut_file(self, file, sep):
        # some spectronaut files include european decimal separators
        if isinstance(file, pd.DataFrame):
            df = file
        else:
            df = pd.read_csv(file, sep=sep, low_memory=False)

            if df[self.intensity_column].dtype != np.float64:
                # load european
                df = pd.read_csv(file, sep=sep, decimal=",")

        self.rawinput = df
        
    

#filter_with_Qvalue	
#TRUE(default) will filter out the intensities that have greater than qvalue_cutoff in EG.Qvalue column. Those intensities will be replaced with zero and will be considered as censored missing values for imputation purpose.

#qvalue_cutoff	
#Cutoff for EG.Qvalue. default is 0.01.

# Protein Level
# PG.Quantity
# PG.ProteinGroups

# Peptide Level
# F.PeakArea
# PEP.StrippedSequence




import os
import sys
from streamlit.web import cli as stcli

def run():
    file_path = os.path.realpath(__file__)
    os.chdir(os.path.dirname(file_path))
    os.system("python -m streamlit run AlphaPeptStats.py --global.developmentMode=false --server.maxUploadSize 500")
    _this_file = os.path.abspath(__file__)
    _this_directory = os.path.dirname(_this_file)

    file_path = os.path.join(_this_directory, 'AlphaPeptStats.py')

    HOME = os.path.expanduser("~")
    ST_PATH = os.path.join(HOME, ".streamlit")

    for folder in [ST_PATH]:
        if not os.path.isdir(folder):
            os.mkdir(folder)


    print(f'Starting AlphaPeptStats from {file_path}')

    args = ["streamlit", "run", file_path, "--global.developmentMode=false", "--browser.gatherUsageStats=False", "--server.maxUploadSize 500"]

    sys.argv = args

    sys.exit(stcli.main())



import pandas as pd
import logging
import streamlit as st
from datetime import datetime
from alphastats.plots.VolcanoPlot import VolcanoPlot


def check_if_options_are_loaded(f):
    # decorator to check for missing values
    def inner(*args, **kwargs):
        if hasattr(st.session_state, "plotting_options") is False:
            load_options()

        return f(*args, **kwargs)

    return inner


def read_uploaded_file_into_df(file):
    filename = file.name

    if filename.endswith(".xlsx"):
        df = pd.read_excel(file)

    elif filename.endswith(".txt") or filename.endswith(".tsv"):
        df = pd.read_csv(file, delimiter="\t")

    elif filename.endswith(".csv"):
        df = pd.read_csv(file)

    else:
        df = None
        logging.warning(
            "WARNING: File could not be read. \nFile has to be a .xslx, .tsv, .csv or .txt file"
        )
        return

    return df


def get_unique_values_from_column(column):
    unique_values = st.session_state.dataset.metadata[column].unique().tolist()
    return unique_values


def st_general(method_dict):

    chosen_parameter_dict = {}

    if "settings" in list(method_dict.keys()):

        settings_dict = method_dict.get("settings")

        for parameter in settings_dict:

            parameter_dict = settings_dict[parameter]

            if "options" in parameter_dict.keys():
                chosen_parameter = st.selectbox(
                    parameter_dict.get("label"), options=parameter_dict.get("options")
                )
            else:
                chosen_parameter = st.checkbox(parameter_dict.get("label"))

            chosen_parameter_dict[parameter] = chosen_parameter

    submitted = st.button("Submit")

    if submitted:
        with st.spinner("Calculating..."):
            return method_dict["function"](**chosen_parameter_dict)


@st.cache_data
def gui_volcano_plot_differential_expression_analysis(
        chosen_parameter_dict, user_session_id = st.session_state.user_session_id
    ):
    """
    initalize volcano plot object with differential expression analysis results
    """
    volcano_plot = VolcanoPlot(
        dataset=st.session_state.dataset, 
        **chosen_parameter_dict, 
        plot = False
    )
    volcano_plot._perform_differential_expression_analysis()
    volcano_plot._add_hover_data_columns()
    return volcano_plot

def gui_volcano_plot():
    """
    Draw Volcano Plot using the VolcanoPlot class
    """
    chosen_parameter_dict = helper_compare_two_groups()
    method = st.selectbox(
        "Differential Analysis using:",
        options=["ttest", "anova", "wald", "sam", "paired-ttest", "welch-ttest"],
    )
    chosen_parameter_dict.update({"method": method})

    # TODO streamlit doesnt allow nested columns check for updates
    
    labels = st.checkbox("Add label")

    draw_line = st.checkbox("Draw line")

    alpha = st.number_input(
            label="alpha", min_value=0.001, max_value=0.050, value=0.050
        )

    min_fc = st.select_slider("Foldchange cutoff", range(0, 3), value=1)
 
    plotting_parameter_dict = {
            "labels": labels,
            "draw_line": draw_line,
            "alpha": alpha,
            "min_fc": min_fc,
        }

    if method == "sam":
        perm = st.number_input(
            label="Number of Permutations", min_value=1, max_value=1000, value=10
        )
        fdr = st.number_input(
            label="FDR cut off", min_value=0.005, max_value=0.1, value=0.050
        )
        chosen_parameter_dict.update({"perm": perm, "fdr": fdr})

   
    submitted = st.button("Submit")

    if submitted:
        volcano_plot = gui_volcano_plot_differential_expression_analysis(
            chosen_parameter_dict, user_session_id = st.session_state.user_session_id
        )
        volcano_plot._update(plotting_parameter_dict)
        volcano_plot._annotate_result_df()
        volcano_plot._plot()
        return volcano_plot.plot


def get_analysis_options_from_dict(method, options_dict):
    """
    extract plotting options from dict amd display as selectbox or 
    give selceted options to plotting function
    """

    method_dict = options_dict.get(method)

    if method == "t-SNE Plot":
        return st_tsne_options(method_dict)

    elif method == "Differential Expression Analysis - T-test":
        return st_calculate_ttest(method=method, options_dict=options_dict)

    elif method == "Differential Expression Analysis - Wald-test":
        return st_calculate_waldtest(method=method, options_dict=options_dict)

    elif method == "Volcano Plot":
        return gui_volcano_plot()

    elif method == "PCA Plot":
        return st_plot_pca(method_dict)

    elif method == "UMAP Plot":
        return st_plot_umap(method_dict)

    elif "settings" not in method_dict.keys():

        if st.session_state.dataset.mat.isna().values.any() == True:
            st.error(
                "Data contains missing values impute your data before plotting (Preprocessing - Imputation)."
            )
            return

        chosen_parameter_dict = {}
        return method_dict["function"](**chosen_parameter_dict)

    else:
        return st_general(method_dict=method_dict)


def st_plot_pca(method_dict):
    chosen_parameter_dict = helper_plot_dimensionality_reduction(
        method_dict=method_dict
    )

    submitted = st.button("Submit")

    if submitted:
        with st.spinner("Calculating..."):
            return method_dict["function"](**chosen_parameter_dict)


def st_plot_umap(method_dict):
    chosen_parameter_dict = helper_plot_dimensionality_reduction(
        method_dict=method_dict
    )

    submitted = st.button("Submit")

    if submitted:
        with st.spinner("Calculating..."):
            return method_dict["function"](**chosen_parameter_dict)


def st_calculate_ttest(method, options_dict):
    """
    perform ttest in streamlit
    """
    chosen_parameter_dict = helper_compare_two_groups()
    chosen_parameter_dict.update({"method": "ttest"})

    submitted = st.button("Submit")

    if submitted:
        with st.spinner("Calculating..."):
            return options_dict.get(method)["function"](**chosen_parameter_dict)


def st_calculate_waldtest(method, options_dict):
    chosen_parameter_dict = helper_compare_two_groups()
    chosen_parameter_dict.update({"method": "wald"})

    submitted = st.button("Submit")

    if submitted:
        with st.spinner("Calculating..."):
            return options_dict.get(method)["function"](**chosen_parameter_dict)


def helper_plot_dimensionality_reduction(method_dict):
    group = st.selectbox(
        method_dict["settings"]["group"].get("label"),
        options=method_dict["settings"]["group"].get("options"),
    )

    circle = False

    if group is not None:

        circle = st.checkbox("circle")

    chosen_parameter_dict = {
        "circle": circle,
        "group": group,
    }
    return chosen_parameter_dict


def helper_compare_two_groups():
    """
    Helper function to compare two groups for example
    Volcano Plot, Differetial Expression Analysis and t-test
    selectbox based on selected column
    """

    chosen_parameter_dict = {}
    group = st.selectbox(
        "Grouping variable",
        options=["< None >"] + st.session_state.dataset.metadata.columns.to_list(),
    )

    if group != "< None >":

        unique_values = get_unique_values_from_column(group)

        group1 = st.selectbox("Group 1", options=unique_values)

        group2 = st.selectbox("Group 2", options=list(reversed(unique_values)))

        chosen_parameter_dict.update(
            {"column": group, "group1": group1, "group2": group2}
        )

        if group1 == group2:
            st.error(
                "Group 1 and Group 2 can not be the same please select different group."
            )

    else:

        group1 = st.multiselect(
                "Group 1 samples:",
                options=st.session_state.dataset.metadata[st.session_state.dataset.sample].to_list(),
            )

        group2 = st.multiselect(
                "Group 2 samples:",
                options=list(
                    reversed(st.session_state.dataset.metadata[st.session_state.dataset.sample].to_list())
                ),
            )

        intersection_list = list(set(group1).intersection(set(group2)))

        if len(intersection_list) > 0:
            st.warning(
                "Group 1 and Group 2 contain same samples: " + str(intersection_list)
            )

        chosen_parameter_dict.update({"group1": group1, "group2": group2})

    return chosen_parameter_dict


def get_sample_names_from_software_file():
    """
    extract sample names from software
    """
    regex_find_intensity_columns = st.session_state.loader.intensity_column.replace(
        "[sample]", ".*"
    )

    df = st.session_state.loader.rawinput
    df = df.set_index(st.session_state.loader.index_column)
    df = df.filter(regex=(regex_find_intensity_columns), axis=1)
    # remove Intensity so only sample names remain
    substring_to_remove = regex_find_intensity_columns.replace(".*", "")
    df.columns = df.columns.str.replace(substring_to_remove, "")
    return df.columns.to_list()


def get_analysis(method, options_dict):

    if method in options_dict.keys():
        obj = get_analysis_options_from_dict(method, options_dict=options_dict)
        return obj


def st_tsne_options(method_dict):
    chosen_parameter_dict = helper_plot_dimensionality_reduction(
        method_dict=method_dict
    )

    n_iter = st.select_slider(
        "Maximum number of iterations for the optimization",
        range(250, 2001),
        value=1000,
    )
    perplexity = st.select_slider("Perplexity", range(5, 51), value=30)

    submitted = st.button("Submit")
    chosen_parameter_dict.update(
        {
            "n_iter": n_iter,
            "perplexity": perplexity,
        }
    )

    if submitted:
        with st.spinner("Calculating..."):
            return method_dict["function"](**chosen_parameter_dict)


def load_options():

    from alphastats.gui.utils.options import plotting_options, statistic_options

    st.session_state["plotting_options"] = plotting_options
    st.session_state["statistic_options"] = statistic_options



import plotly.express as px
import requests
import pandas as pd
from io import StringIO
import numpy as np
from alphastats.utils import check_internetconnection, check_if_df_empty


class enrichement_df(pd.DataFrame):
    # this is that added methods dont get lost when operatons on pd Dataframe get performed
    @property
    def _constructor(self):
        return enrichement_df

    def _modify_df(self):
        self["Description"] = self["term"] + " " + self["description"]
        self["-log10(P-value)"] = -np.log10(self["p_value"])
        self["over_under"] = np.where(
            self["over_under"] == "o", "over-represented", "under-represented"
        )

    @check_if_df_empty
    def plot_scatter(self):
        """
        Plot Scatterplot with -log10(p-value) on x-axis and effect size on y-axis.
        Datapoints will be colored by Gene Ontology-term category.

        Returns:
            plotly.graph_objects._figure.Figure: Scatterplot of GO terms
        """
        self._modify_df()
        plot = px.scatter(
            self,
            x="-log10(P-value)",
            y="effect_size",
            size=self["foreground_count"],
            color="category",
            hover_name=self["Description"],
        )
        return plot

    @check_if_df_empty
    def plot_bar(self):
        """
        Plot p-values as Barplot

        Returns:
            plotly.graph_objects._figure.Figure: Barplot
        """
        self._modify_df()
        plot = px.bar(
            self,
            x="-log10(P-value)",
            y="Description",
            orientation="h",
            color="over_under",
            color_discrete_map={
                "under-represented": "#009599",
                "over-represented": "#B65EAF",
            },
        )
        return plot


class Enrichment:
    @staticmethod
    def _extract_protein_ids(entry):
        try:
            proteins = entry.split(",")
            protein_id_list = []
            for protein in proteins:
                # 'sp|P0DMV9|HS71B_HUMAN,sp|P0DMV8|HS71A_HUMAN',
                if "|" in protein:
                    fasta_header_split = protein.split("|")
                else:
                    fasta_header_split = protein
                if isinstance(fasta_header_split, str):
                    #  'ENSEMBL:ENSBTAP00000007350',
                    if "ENSEMBL:" in fasta_header_split:
                        protein_id = fasta_header_split.replace("ENSEMBL:", "")
                    else:
                        protein_id = fasta_header_split
                else:
                    protein_id = fasta_header_split[1]
                protein_id_list.append(protein_id)
            protein_id_concentate = ";".join(protein_id_list)
            # ADD REV to the protein ID, else there will be duplicates in the ProteinGroup column
            if "REV_" in entry:
                protein_id_concentate = "REV_" + protein_id_concentate

        except AttributeError:
            protein_id_concentate = entry

        return protein_id_concentate

    def _get_ptm_proteins(self, sample=None):

        if self.evidence_df is None:
            raise ValueError(
                "No informations about PTMs."
                "Either load a list of ProteinIDs containing PTMs"
                "or DataSet.load_ptm_df()"
            )

        if "ProteinGroup" not in self.evidence_df.columns:
            self.evidence_df["ProteinGroup"] = self.evidence_df["Proteins"].map(
                self._extract_protein_ids
            )

        if isinstance(sample, str):
            protein_list = self.evidence_df[
                (self.evidence_df["Modifications"] != "Unmodified")
                & (self.evidence_df["Experiment"] == sample)
            ]["ProteinGroup"].to_list()

        elif isinstance(sample, list):
            protein_list = self.evidence_df[
                (self.evidence_df["Modifications"] != "Unmodified")
                & (self.evidence_df["Experiment"].isin(sample))
            ]["ProteinGroup"].to_list()

        else:
            protein_list = self.evidence_df[
                self.evidence_df["Modifications"] != "Unmodified"
            ]["ProteinGroup"].to_list()

        protein_list = [str(x) for x in protein_list]
        return protein_list

    def go_characterize_foreground(self, protein_list, tax_id=9606):
        """
        Display existing functional annotations for your protein(s) of interest.
        No statistical test for enrichment is performed.
        Using the API connection from a GO tool: https://agotool.org

        Args:
            tax_id (int, optional): NCBI taxon identifier used as background. Defaults to 9606 (=Homo sapiens).
            protein_list (list): list of enriched protein ids in the foreground sample.

        Returns:
            pandas.DataFrame: DataFrame
            * ``'rank'``: The rank is a combination of uncorrected p value and effect size (based on s value). It serves to highlight the most interesting results and tries to emphasize the importance of the effect size.
            * ``'term'``: A unique identifier for a specific functional category.
            * ``'description'``: A short description (or title) of a functional term.
            * ``'p value corrected'``: p value without multiple testing correction, stemming from either Fisher's exact test or Kolmorov Smirnov test (only for "Gene Ontology Cellular Component TEXTMINING", "Brenda Tissue Ontoloy", and "Disease Ontology" since these are based on a continuous score from text mining rather than a binary classification).
            * ``'effect size'``: Proportion of the Foregrounda nd the Background
            * ``'description'``: A short description (or title) of a functional term.
            * ``'year'``: Year of the scientific publication.
            * ``'over_under'``: Overrepresented (o) or underrepresented (u).
            * ``'s_value'``: The s value is a combination of (minus log) p value and effect size.
            * ``'ratio_in_foreground'``: The ratio in the ForeGround is calculated by dividing the number of positive associations for a given term by the number of input proteins (protein groups) for the Foreground.
            * ``'ratio_in_background'``: The ratio in the BackGround is analogous to the above ratio in the FG, using the associations for the background and Background input proteins instead.
            * ``'foreground_count'``: The ForeGround count consists of the number of all positive associations for the given term (i.e. how many proteins are associated with the given term).
            * ``'foreground_n'``: ForeGround n is comprised of the number of input proteins for the Foreground.
            * ``'background_count'``: The BackGround count is analogous to the "FG count" for the Background.
            * ``'background_n'``: BackGround n is analogous to "FG n".
            * ``'foreground_ids'``: ForeGround IDentifierS are semicolon separated protein identifers of the Forground that are associated with the given term.
            * ``'background_ids'``: BackGround IDentifierS are analogous to "FG IDs" for the Background.
            * ``'etype'``: Short for "Entity type", numeric internal identifer for different functional categories.

        """
        check_internetconnection()
        protein_list = [self._extract_protein_ids(protein) for protein in protein_list]
        protein_list = "%0d".join(protein_list)
        url = r"https://agotool.org/api_orig"

        result = requests.post(
            url,
            params={
                "output_format": "tsv",
                "enrichment_method": "characterize_foreground",
                "taxid": tax_id,
            },
            data={"foreground": protein_list},
        )

        result_df = enrichement_df(pd.read_csv(StringIO(result.text), sep="\t"))
        return result_df

    def go_abundance_correction(self, bg_sample, fg_sample=None, fg_protein_list=None):
        """
        Gene Ontology Enrichement Analysis with abundance correction.
        Using the API connection from a GO tool: https://agotool.org

        For the analysis modified proteins in the foreground sample are compared with proteins and
        their intensity of the background sample.
        In case there is no information about PTMs in the dataset a list of enriched proteins in the
        foreground can be loaded. This list of Protein IDs can be obtaint by performing a differential
        expression analysis or a ANOVA.


        Args:
            fg_sample (str): name of foreground sample
            bg_sample (str): name of background sample
            fg_protein_list (list, optional): list of enriched protein ids in the foreground sample. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame
            * ``'rank'``: The rank is a combination of uncorrected p value and effect size (based on s value). It serves to highlight the most interesting results and tries to emphasize the importance of the effect size.
            * ``'term'``: A unique identifier for a specific functional category.
            * ``'description'``: A short description (or title) of a functional term.
            * ``'p value corrected'``: p value without multiple testing correction, stemming from either Fisher's exact test or Kolmorov Smirnov test (only for "Gene Ontology Cellular Component TEXTMINING", "Brenda Tissue Ontoloy", and "Disease Ontology" since these are based on a continuous score from text mining rather than a binary classification).
            * ``'effect size'``: Proportion of the Foregrounda nd the Background
            * ``'description'``: A short description (or title) of a functional term.
            * ``'year'``: Year of the scientific publication.
            * ``'over_under'``: Overrepresented (o) or underrepresented (u).
            * ``'s_value'``: The s value is a combination of (minus log) p value and effect size.
            * ``'ratio_in_foreground'``: The ratio in the ForeGround is calculated by dividing the number of positive associations for a given term by the number of input proteins (protein groups) for the Foreground.
            * ``'ratio_in_background'``: The ratio in the BackGround is analogous to the above ratio in the FG, using the associations for the background and Background input proteins instead.
            * ``'foreground_count'``: The ForeGround count consists of the number of all positive associations for the given term (i.e. how many proteins are associated with the given term).
            * ``'foreground_n'``: ForeGround n is comprised of the number of input proteins for the Foreground.
            * ``'background_count'``: The BackGround count is analogous to the "FG count" for the Background.
            * ``'background_n'``: BackGround n is analogous to "FG n".
            * ``'foreground_ids'``: ForeGround IDentifierS are semicolon separated protein identifers of the Forground that are associated with the given term.
            * ``'background_ids'``: BackGround IDentifierS are analogous to "FG IDs" for the Background.
            * ``'etype'``: Short for "Entity type", numeric internal identifer for different functional categories.
        """

        check_internetconnection()
        # get PTMs for fg_sample
        if fg_protein_list is None:
            fg_protein_list = self._get_ptm_proteins(sample=fg_sample)

        fg_protein_list = [
            self._extract_protein_ids(protein) for protein in fg_protein_list
        ]
        fg_protein_list = "%0d".join(fg_protein_list)

        # get intensity for bg_sample
        bg_protein = "%0d".join(self.mat.loc[bg_sample].index.to_list())
        bg_intensity = "%0d".join(self.mat.loc[bg_sample].astype(str).values.tolist())

        url = r"https://agotool.org/api_orig"
        result = requests.post(
            url,
            params={
                "output_format": "tsv",
                "enrichment_method": "abundance_correction",
            },
            data={
                "foreground": fg_protein_list,
                "background": bg_protein,
                "background_intensity": bg_intensity,
            },
        )
        result_df = enrichement_df(pd.read_csv(StringIO(result.text), sep="\t"))
        return result_df

    def go_compare_samples(self, fg_sample, bg_sample):
        """
        Gene Ontology Enrichement Analysis without abundance correction.
        Using the API connection from a GO tool: https://agotool.org

        Args:
            fg_sample (str): name of the foreground sample
            bg_sample (str): name of the background sample

        Returns:
            pandas.DataFrame: DataFrame
            * ``'rank'``: The rank is a combination of uncorrected p value and effect size (based on s value). It serves to highlight the most interesting results and tries to emphasize the importance of the effect size.
            * ``'term'``: A unique identifier for a specific functional category.
            * ``'description'``: A short description (or title) of a functional term.
            * ``'p value corrected'``: p value without multiple testing correction, stemming from either Fisher's exact test or Kolmorov Smirnov test (only for "Gene Ontology Cellular Component TEXTMINING", "Brenda Tissue Ontoloy", and "Disease Ontology" since these are based on a continuous score from text mining rather than a binary classification).
            * ``'effect size'``: Proportion of the Foregrounda nd the Background
            * ``'description'``: A short description (or title) of a functional term.
            * ``'year'``: Year of the scientific publication.
            * ``'over_under'``: Overrepresented (o) or underrepresented (u).
            * ``'s_value'``: The s value is a combination of (minus log) p value and effect size.
            * ``'ratio_in_foreground'``: The ratio in the ForeGround is calculated by dividing the number of positive associations for a given term by the number of input proteins (protein groups) for the Foreground.
            * ``'ratio_in_background'``: The ratio in the BackGround is analogous to the above ratio in the FG, using the associations for the background and Background input proteins instead.
            * ``'foreground_count'``: The ForeGround count consists of the number of all positive associations for the given term (i.e. how many proteins are associated with the given term).
            * ``'foreground_n'``: ForeGround n is comprised of the number of input proteins for the Foreground.
            * ``'background_count'``: The BackGround count is analogous to the "FG count" for the Background.
            * ``'background_n'``: BackGround n is analogous to "FG n".
            * ``'foreground_ids'``: ForeGround IDentifierS are semicolon separated protein identifers of the Forground that are associated with the given term.
            * ``'background_ids'``: BackGround IDentifierS are analogous to "FG IDs" for the Background.
            * ``'etype'``: Short for "Entity type", numeric internal identifer for different functional categories.
        """

        check_internetconnection()
        # get protein ids for samples
        fg_proteins = "%0d".join(self._get_ptm_proteins(sample=fg_sample))
        bg_proteins = "%0d".join(self._get_ptm_proteins(sample=bg_sample))

        url = r"https://agotool.org/api_orig"
        result = requests.post(
            url,
            params={"output_format": "tsv", "enrichment_method": "compare_samples"},
            data={"foreground": fg_proteins, "background": bg_proteins},
        )
        result_df = enrichement_df(pd.read_csv(StringIO(result.text), sep="\t"))
        return result_df

    def go_genome(self, tax_id=9606, fg_sample=None, protein_list=None):
        """
        Gene Ontology Enrichement Analysis using a Background from UniProt Reference Proteomes.
        Using the API connection from a GO tool: https://agotool.org

        Args:
            tax_id (int, optional): NCBI taxon identifier used as background. Defaults to 9606 (=Homo sapiens).
            fg_sample (str, optional): name of sample used as foreground. Defaults to None.
            protein_list (list, optional): list of enriched protein ids in the foreground sample. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame
            * ``'rank'``: The rank is a combination of uncorrected p value and effect size (based on s value). It serves to highlight the most interesting results and tries to emphasize the importance of the effect size.
            * ``'term'``: A unique identifier for a specific functional category.
            * ``'description'``: A short description (or title) of a functional term.
            * ``'p value corrected'``: p value without multiple testing correction, stemming from either Fisher's exact test or Kolmorov Smirnov test (only for "Gene Ontology Cellular Component TEXTMINING", "Brenda Tissue Ontoloy", and "Disease Ontology" since these are based on a continuous score from text mining rather than a binary classification).
            * ``'effect size'``: Proportion of the Foregrounda nd the Background
            * ``'description'``: A short description (or title) of a functional term.
            * ``'year'``: Year of the scientific publication.
            * ``'over_under'``: Overrepresented (o) or underrepresented (u).
            * ``'s_value'``: The s value is a combination of (minus log) p value and effect size.
            * ``'ratio_in_foreground'``: The ratio in the ForeGround is calculated by dividing the number of positive associations for a given term by the number of input proteins (protein groups) for the Foreground.
            * ``'ratio_in_background'``: The ratio in the BackGround is analogous to the above ratio in the FG, using the associations for the background and Background input proteins instead.
            * ``'foreground_count'``: The ForeGround count consists of the number of all positive associations for the given term (i.e. how many proteins are associated with the given term).
            * ``'foreground_n'``: ForeGround n is comprised of the number of input proteins for the Foreground.
            * ``'background_count'``: The BackGround count is analogous to the "FG count" for the Background.
            * ``'background_n'``: BackGround n is analogous to "FG n".
            * ``'foreground_ids'``: ForeGround IDentifierS are semicolon separated protein identifers of the Forground that are associated with the given term.
            * ``'background_ids'``: BackGround IDentifierS are analogous to "FG IDs" for the Background.
            * ``'etype'``: Short for "Entity type", numeric internal identifer for different functional categories.
        """

        check_internetconnection()

        if protein_list is None:
            protein_list = self._get_ptm_proteins(sample=fg_sample)

        protein_list = [self._extract_protein_ids(protein) for protein in protein_list]
        protein_list = "%0d".join(protein_list)
        url = r"https://agotool.org/api_orig"

        result = requests.post(
            url,
            params={
                "output_format": "tsv",
                "enrichment_method": "genome",
                "taxid": tax_id,
            },
            data={"foreground": protein_list},
        )

        result_df = enrichement_df(pd.read_csv(StringIO(result.text), sep="\t"))
        return result_df



import functools
import warnings
from typing import Type
import logging
import http.client as httplib


def ignore_warning(warning: Type[Warning]):
    """
    Ignore a given warning occurring during method execution.

    Args:
        warning (Warning): warning type to ignore.

    Returns:
        the inner function

    """

    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=warning)
                return func(*args, **kwargs)

        return wrapper

    return inner


class LoaderError(Exception):
    """Loader Exception that will be logged."""

    def __init__(self, message):
        self.message = message
        logging.error(message)
        super().__init__(self.message)


def check_for_missing_values(f):
    # decorator to check for missing values
    def inner(*args, **kwargs):
        if args[0].mat.isna().values.any() is True:
            raise ValueError(
                "Data contains missing values. Consider Imputation:"
                "for instance `DataSet.preprocess(imputation='mean')`."
            )
        return f(*args, **kwargs)

    return inner


def check_internetconnection():
    connection = httplib.HTTPSConnection("8.8.8.8", timeout=5)
    try:
        connection.request("HEAD", "/")
        return True
    except Exception:
        raise ConnectionError("No internet connection available.")
    finally:
        connection.close()


def check_if_df_empty(f):
    # decorator to check for missing values
    def inner(*args, **kwargs):
        if args[0].empty is True:
            raise ValueError("DataFrame is empty. No significant GO-terms found.")
        return f(*args, **kwargs)

    return inner

def list_to_tuple(function):
    """
    list are not hashable not suitable for caching 
    convert to tuple
    """
    def wrapper(*args):
        args = [tuple(x) if type(x) == list else x for x in args]
        result = function(*args)
        result = tuple(result) if type(result) == list else result
        return result
    return wrapper

def find_duplicates_in_list(l:list) -> list:
    """Find duplicates in a list

    Args:
        l (list): list

    Returns:
        list: list with duplicated values
    """
    duplicates = []
    ys = sorted(l)

    for n in range(1, len(l)):
        if ys[n] == ys[n-1]:
            duplicates.append(ys[n])
    return duplicates
    


import numpy as np
import pandas as pd
import numba as nb
import numba_stats as nbs
import plotly.express as px
import plotly.graph_objects as go
from sklearn.linear_model import LinearRegression
from scipy import stats
import statsmodels.api as sm
from collections import Counter
import multiprocessing
from joblib import Parallel, delayed
import random
import math
import swifter
from sklearn.preprocessing import StandardScaler
from statsmodels.stats.multitest import multipletests
import os
from pathlib import Path

# code taken from Isabel Bludau - multicova

def get_std(x):
    """
    Function to calculate the sample standard deviation.
    """
    std_x = np.sqrt(np.sum((abs(x - x.mean())**2)/(len(x)-1)))
    return std_x


def perform_ttest(x, y, s0):
    """
    Function to perform a independent two-sample t-test including s0
    adjustment.
    Assumptions: Equal or unequal sample sizes with similar variances.
    """
    mean_x = np.mean(x)
    mean_y = np.mean(y)
    # Get fold-change
    fc = mean_x-mean_y
    n_x = len(x)
    n_y = len(y)
    # pooled standard vatiation
    # assumes that the two distributions have the same variance
    sp = np.sqrt((((n_x-1)*get_std(x)**2) + ((n_y-1)*(get_std(y)**2)))/(n_x+n_y-2))
    # Get t-values
    tval = fc/(sp * (np.sqrt((1/n_x)+(1/n_y))))
    tval_s0 = fc/((sp * (np.sqrt((1/n_x)+(1/n_y))))+s0)
    # Get p-values
    pval = 2*(1-stats.t.cdf(np.abs(tval), n_x+n_y-2))
    pval_s0 = 2*(1-stats.t.cdf(np.abs(tval_s0), n_x+n_y-2))
    return [fc, tval, pval, tval_s0, pval_s0]


def permutate_vars(X, n_rand, seed=42):
    """
    Function to randomly permutate an array `n_rand` times.
    """
    x_perm_idx = generate_perms(n=len(X), n_rand=n_rand, seed=seed)
    X_rand = list()
    for i in np.arange(0, len(x_perm_idx)):
        X_rand.append([X[j] for j in x_perm_idx[i]])
    return X_rand


def workflow_ttest(df, c1, c2, s0=1, parallelize=False):
    """
    Function to perform a t-test on all rows in a data frame.
    c1 specifies the column names with intensity values of the first condition.
    c2 specifies the column names with intensity values of the second
    condition. s0 is the tuning parameter that specifies the minimum
    fold-change to be trusted.
    """
    if parallelize:
        res = df.swifter.progress_bar(False).apply(lambda row : perform_ttest(row[c1], row[c2], s0=s0), axis = 1)
    else:
        res = df.apply(lambda row : perform_ttest(row[c1], row[c2], s0=s0), axis = 1)

    res = pd.DataFrame(list(res), columns=['fc','tval','pval','tval_s0','pval_s0'])

    df_real = pd.concat([df, res], axis=1)

    return df_real


def workflow_permutation_tvals(df, c1, c2, s0=1, n_perm=2, parallelize=False):
    """
    Function to perform a t-test on all rows in a data frame based on the
    permutation of samples across conditions.
    c1 specifies the column names with intensity value sof the first condition.
    c2 specifies the column names with intensity value sof the second
    condition. s0 is the tuning parameter that specifies the minimum
    fold-change to be trusted. n_perm specifies the number of random
    permutations to perform.
    """
    all_c = c1 + c2
    all_c_rand = permutate_vars(all_c, n_rand=n_perm)
    res_perm = list()
    for i in np.arange(0,len(all_c_rand)):
        if parallelize:
            res_i = df.swifter.progress_bar(False).apply(lambda row : perform_ttest(row[all_c_rand[i][0:len(c1)]],
                                                                row[all_c_rand[i][len(c1):len(c1)+len(c2)]],
                                                                s0=s0),
                                                                axis = 1)
        else:
            res_i = df.apply(lambda row : perform_ttest(row[all_c_rand[i][0:len(c1)]],
                                                        row[all_c_rand[i][len(c1):len(c1)+len(c2)]],
                                                        s0=s0),
                                                        axis=1)
        res_i = pd.DataFrame(list(res_i), columns=['fc','tval','pval','tval_s0','pval_s0'])
        res_perm.append(list(np.sort(np.abs(res_i.tval_s0.values))))
    return res_perm


def get_tstat_cutoff(res_real, t_perm_avg, delta):
    # Extract all t-stats from the results of the 'real' data
    # and sort the absolute values.
    t_real = res_real.tval_s0
    t_real_abs = np.sort(np.abs(t_real))
    # print(t_real_abs)

    # Calculate the difference between the observed t-stat
    # and the average random t-stat for each rank position.
    t_diff = t_real_abs - t_perm_avg
    # print(t_diff)

    # Find the minimum t-stat value for which the difference
    # between the observed and average random t-stat is
    # larger than the selected delta.
    t_max = t_real_abs[t_diff > delta]
    if (t_max.shape[0] == 0):
        t_max = np.ceil(np.max(t_real_abs))
    else:
        t_max = np.min(t_max)
    return t_max


@nb.njit
def get_positive_count(res_real_tval_s0, t_cut):
    """
    Count number of tval_s0 in res_real that are above the t_cut threshold.
    """
    ts_sig = list()
    for r in res_real_tval_s0:
        r_abs = np.abs(r)
        if r_abs >= t_cut:
            ts_sig.append(r_abs)

    return len(ts_sig)


@nb.njit
def get_false_positive_count(res_perm, t_cut):
    """
    Get median number of tval_s0 in res_perm that are above the t_cut
    threshold.
    """
    ts_sig = list()
    for rp in res_perm:
        ts_abs = np.abs(rp)
        ts_max = list()
        for t in ts_abs:
            if t >= t_cut:
                ts_max.append(t)

        ts_sig.append(len(ts_max))

    res = np.median(np.array(ts_sig))

    return res


def get_pi0(res_real, res_perm):
    """
    Estimate pi0, the proportion of true null (unaffected) genes in the data
    set, as follows:
    (a) Compute q25; q75 = 25% and 75% points of the permuted d values
    (if p = # genes, B = # permutations, there are pB such d values).
    (b) Count the number of values d in the real dataset that are between q25
    and q75 (there are p values to test). pi0 = #d/0.5*p.
    (c) Let pi0 = min(pi0; 1) (i.e., truncate at 1).
    Documentation in: https://statweb.stanford.edu/~tibs/SAM/sam.pdf
    """
    t_real = res_real["tval_s0"]
    t_real_abs = np.sort(np.abs(t_real))

    t_perm = list()
    for ri in res_perm:
        for r in ri:
            t_perm.append(r)
    t_perm_abs = np.sort(np.abs(t_perm))

    t_perm_25 = np.percentile(t_perm_abs, 25)
    t_perm_75 = np.percentile(t_perm_abs, 75)

    n_real_in_range = np.sum((t_real_abs >= t_perm_25) & (t_real_abs <= t_perm_75))
    pi0 = n_real_in_range/(0.5*len(t_real_abs))

    # pi0 can maximally be 1
    pi0 = np.min([pi0, 1])

    return pi0


def get_fdr(n_pos, n_false_pos, pi0):
    """
    Compute the FDR by dividing the number of false positives by the number of
    true positives. The number of false positives are adjusted by pi0, the
    proportion of true null (unaffected) genes in the data set.
    """
    n = n_false_pos*pi0
    if n != 0:
        if n_pos != 0:
            fdr = n/n_pos
        else:
            fdr = 0
    else:
        if n_pos > 0:
            fdr = 0
        else:
            fdr = np.nan
    return fdr


def estimate_fdr_stats(res_real, res_perm, delta):
    """
    Helper function for get_fdr_stats_across_deltas.
    It computes the FDR and tval_s0 thresholds for a specified delta.
    The function returns a list of the following values:
    t_cut, n_pos, n_false_pos, pi0, n_false_pos_corr, fdr
    """
    perm_avg = np.mean(res_perm, axis=0)
    t_cut = get_tstat_cutoff(res_real, perm_avg, delta)
    n_pos = get_positive_count(res_real_tval_s0=np.array(res_real.tval_s0), t_cut=t_cut)
    n_false_pos = get_false_positive_count(np.array(res_perm), t_cut=t_cut)
    pi0 = get_pi0(res_real, res_perm)
    n_false_pos_corr = n_false_pos*pi0
    fdr = get_fdr(n_pos, n_false_pos, pi0)
    return [t_cut, n_pos, n_false_pos, pi0, n_false_pos_corr, fdr]


def get_fdr_stats_across_deltas(res_real, res_perm):
    """
    Wrapper function that starts with the res_real and res_perm to derive
    the FDR and tval_s0 thresholds for a range of different deltas.
    """
    res_stats = list()
    for d in np.arange(0, 10, 0.01):
        res_d = estimate_fdr_stats(res_real, res_perm, d)
        if np.isnan(res_d[5]):
            break
        else:
            res_stats.append(res_d)
    res_stats_df = pd.DataFrame(res_stats,
                                columns=['t_cut', 'n_pos', 'n_false_pos', 'pi0', 'n_false_pos_corr', 'fdr'])

    return(res_stats_df)


def get_tstat_limit(stats, fdr=0.01):
    """
    Function to get tval_s0 at the specified FDR.
    """
    t_limit = np.min(stats[stats.fdr <= fdr].t_cut)
    return(t_limit)


def annotate_fdr_significance(res_real, stats, fdr=0.01):
    t_limit = np.min(stats[stats.fdr <= fdr].t_cut)
    res_real['qval'] = [np.min(stats[stats.t_cut <= abs(x)].fdr) for x in res_real['tval_s0']]
    res_real['FDR' + str(int(fdr*100)) + '%'] = ["sig" if abs(x) >= t_limit else "non_sig" for x in res_real['tval_s0']]
    return(res_real)


def perform_ttest_getMaxS(fc, s, s0, n_x, n_y):
    """
    Helper function to get ttest stats for specified standard errors s.
    Called from within get_fdr_line.
    """
    # Get t-values
    tval = fc/s
    tval_s0 = fc/(s+s0)

    # Get p-values
    pval = 2*(1-stats.t.cdf(np.abs(tval), n_x+n_y-2))
    pval_s0 = 2*(1-stats.t.cdf(np.abs(tval_s0), n_x+n_y-2))

    return [fc, tval, pval, tval_s0, pval_s0]


def get_fdr_line(t_limit, s0, n_x, n_y, plot=False,
                 fc_s=np.arange(0, 6, 0.01), s_s=np.arange(0.005, 6, 0.005)):
    """
    Function to get the fdr line for a volcano plot as specified tval_s0
    limit, s0, n_x and n_y.
    """
    pvals = np.ones(len(fc_s))
    svals = np.zeros(len(fc_s))
    for i in np.arange(0, len(fc_s)):
        for j in np.arange(0, len(s_s)):
            res_s = perform_ttest_getMaxS(fc=fc_s[i], s=s_s[j], s0=s0, n_x=n_x, n_y=n_y)
            if res_s[3] >= t_limit:
                if svals[i] < s_s[j]:
                    svals[i] = s_s[j]
                    pvals[i] = res_s[2]

    s_df = pd.DataFrame(np.array([fc_s, svals, pvals]).T, columns=['fc_s','svals','pvals'])
    s_df = s_df[s_df.pvals != 1]

    s_df_neg = s_df.copy()
    s_df_neg.fc_s = -s_df_neg.fc_s

    s_df = pd.concat([s_df, s_df_neg])

    if (plot):
        fig = px.scatter(x=s_df.fc_s,
                         y=-np.log10(s_df.pvals),
                         template='simple_white')
        fig.show()

    return(s_df)



def perform_ttest_analysis(df, c1, c2, s0=1, n_perm=2, fdr=0.01, id_col='Genes', plot_fdr_line=False, parallelize=False):
    """
    Workflow function for the entire T-test analysis including FDR
    estimation and visualizing a volcanoe plot.
    """
    ttest_res = workflow_ttest(df, c1, c2, s0, parallelize=parallelize)
    ttest_perm_res = workflow_permutation_tvals(df, c1, c2, s0, n_perm, parallelize=parallelize)
    ttest_stats = get_fdr_stats_across_deltas(ttest_res, ttest_perm_res)
    ttest_res = annotate_fdr_significance(res_real=ttest_res, stats=ttest_stats, fdr=fdr)
    t_limit = get_tstat_limit(stats=ttest_stats, fdr=fdr)
    return(ttest_res, t_limit)


def perform_regression(y, X, s0):
    """
    Function to perform multiple linear regression including the s0 parameter.
    """
    # Use LinearRegression from sklearn.linear_model
    lm = LinearRegression()
    lm.fit(X, y)
    betas = np.append(lm.intercept_, lm.coef_)
    predictions = lm.predict(X)

    # Add intercept column to X dataframe
    newX = pd.DataFrame({"Constant": np.ones(len(X))}).join(pd.DataFrame(X))

    # Calculate mean squared error (MSE) of observed y vs. prediction
    # MSE = squared_error/degrees_of_freedom
    # degrees_of_freedom = number_of_samples - number_of_betas (including beta0 = intercept)
    MSE = (sum((y-predictions)**2))/(len(newX)-len(newX.columns))
    # Calculate t-values and p-values
    ts_b, ps_b = get_tstats(newX, betas, MSE)
    # Adjust the MSE by the minimum fold change that we trust (s0)
    MSE_s0 = MSE + s0
    # Calculate t-values and p-values
    ts_b_s0, ps_b_s0 = get_tstats(newX, betas, MSE_s0)
    betas_std = list()
    for i in np.arange(0, len(betas)):
        betas_std.append(betas[i]*(get_std(np.array(newX)[:, i])/get_std(y)))
    betas_std = np.array(betas_std)
    return betas, betas_std, predictions, MSE, ts_b, ps_b, MSE_s0, ts_b_s0, ps_b_s0


def get_tstats(newX, betas, MSE):
    # Calculate standard deviation for each beta
    # This scales the error as estimated by MSE to each beta
    # (error on the betas might be different for different betas depending on
    # the variable)
    var_b = MSE*(np.linalg.inv(np.dot(newX.T, newX)).diagonal())
    sd_b = np.sqrt(var_b)
    # Get t-values
    ts_b = betas/sd_b
    # Get p-values
    #ps_b =[2*(1-stats.t.cdf(np.abs(i),len(newX)-len(newX.columns))) for i in ts_b]
    #ps_b =[2*(1-nbs.t_cdf(np.abs(i),len(newX)-len(newX.columns),0,1)) for i in ts_b]
    ps_b = get_cdf(ts_b, len(newX)-len(newX.columns))
    return ts_b, ps_b


@nb.njit
def get_cdf(ts, df):
    pvals = [0.0][1:]
    for t in ts:
        pvals.append(2*(1-nbs.t_cdf(np.abs(t), df, 0, 1)))
    return (pvals)


def filter_nan(y, X):
    """
    Find nan values in quantitative data y and remove corresponding samples
    from both X and y.
    """
    nan_mask = np.invert(np.isnan(y))
    y_new = y[nan_mask]
    X_new = X[nan_mask]
    return y_new, X_new


def get_min_vars(X):
    """
    Get the minimum number of observations for each covariate.
    """
    min_vars = list()
    for d in np.arange(0, X.shape[1]):
        test_vals = X[:, d]
        n_vars = len(set(test_vals))
        if n_vars > 2:
            min_vars.append(n_vars)
        else:
            test_vals_0 = np.append(test_vals,[0,1]) # append 0 and 1 for counter
            var_count = Counter(test_vals_0)
            min_vars.append(min(var_count.values()))

    return min_vars


def regression_workflow(y, X, s0):
    y_new, X_new = filter_nan(y, X)
    # Removed min_var filtering after standard scaling of variable.
    # @ToDo: Include sanity check that sufficient categories are covered!
    # The following counts the minimum number of observations for each covariate
    #min_vars = get_min_vars(X_new)
    # A min(min_vars) value < 2 means essentially no values for at least one covariate were observed
    # A beta of zero and p-values of 1 are returned for these cases
    #if min(min_vars) < 2:
    #    betas, betas_std, tvals, pvals, tvals_s0, pvals_s0 = 0, 0, 0, 1, 0, 1
    #else:
    #    betas, betas_std, predictions, MSE, tvals, pvals, MSE_s0, tvals_s0, pvals_s0 = perform_regression(np.array(y_new), np.array(X_new), s0)
    betas, betas_std, predictions, MSE, tvals, pvals, MSE_s0, tvals_s0, pvals_s0 = perform_regression(np.array(y_new), np.array(X_new), s0)
    return betas, betas_std, tvals, pvals, tvals_s0, pvals_s0


def regression_workflow_permutation(y, X_rand, s0):
    res_rand = list()
    for X_r in X_rand:
        res_rand.append(regression_workflow(y, X_r, s0))
    # num_cores = multiprocessing.cpu_count()
    # res_rand = Parallel(n_jobs=num_cores-1)(delayed(regression_workflow)(y=y, X=X_r, s0=s0) for X_r in X_rand)
    return res_rand


def get_fdr_line_regression(t_limits, s0, X, plot = False,
                            fc_s = np.arange(0,6,0.01), s_s = np.arange(0.005,6,0.005)):
    """
    Function to get the fdr line for a volcano plot as specified tval_s0 limit, s0, n_x and n_y.
    """
    #pvals = [list(np.ones(len(fc_s)))] * X.shape[1]
    pvals = [list(np.ones(len(fc_s))) for i in range(0,X.shape[1])]
    #print(pvals)
    #svals = [list(np.zeros(len(fc_s)))] * X.shape[1]
    svals = [list(np.zeros(len(fc_s))) for i in range(0,X.shape[1])]
    #print(svals)
    for i in np.arange(0,len(fc_s)):
        for j in np.arange(0,len(s_s)):
            res_s = perform_ttest_getMaxS_regression(fc=fc_s[i], s=s_s[j], s0=s0, X=X)
            for k in np.arange(0,X.shape[1]):
                t_limit = t_limits[k]
                if (res_s[k][3] >= t_limit) and (svals[k][i] < s_s[j]):
                    svals[k][i] = s_s[j]
                    pvals[k][i] = res_s[k][2]
    s_df_list = list()
    for k in np.arange(0, X.shape[1]):
        s_df = pd.DataFrame(np.array([fc_s,svals[k],pvals[k]]).T, columns=['fc_s','svals','pvals'])
        s_df = s_df[s_df.pvals != 1]

        s_df_neg = s_df.copy()
        s_df_neg.fc_s = -s_df_neg.fc_s

        s_df = pd.concat([s_df, s_df_neg])

        if (plot):
            fig = px.scatter(x=s_df.fc_s,
                             y=-np.log10(s_df.pvals),
                             template='simple_white')
            fig.show()

        s_df_list.append(s_df)

    return(s_df_list)


def perform_ttest_getMaxS_regression(fc, s, s0, X):
    """
    Helper function to get ttest stats for specified standard errors s.
    Called from within get_fdr_line_regression.
    """
    newX = pd.DataFrame({"Constant":np.ones(len(X))}).join(pd.DataFrame(X))
    fc=np.repeat(fc,newX.shape[1])
    #print(fc)
    var_b = s*(np.linalg.inv(np.dot(newX.T,newX)).diagonal())
    sd_b = np.sqrt(var_b)
    ts_b = fc/sd_b
    #ps_b =[2*(1-stats.t.cdf(np.abs(i),len(newX)-len(newX.columns))) for i in ts_b]
    ps_b = get_cdf(ts_b, len(newX)-len(newX.columns))
    var_b_s0 = (s+s0)*(np.linalg.inv(np.dot(newX.T,newX)).diagonal())
    sd_b_s0 = np.sqrt(var_b_s0)
    ts_b_s0 = fc/sd_b_s0
    #ps_b_s0 =[2*(1-stats.t.cdf(np.abs(i),len(newX)-len(newX.columns))) for i in ts_b_s0]
    ps_b_s0 = get_cdf(ts_b_s0, len(newX)-len(newX.columns))
    res = [[fc[x], ts_b[x], ps_b[x], ts_b_s0[x], ps_b_s0[x]] for x in np.arange(1,newX.shape[1])]
    return res

def generate_perms(n, n_rand, seed=42):
    """
    Generate n_rand permutations of indeces ranging from 0 to n.
    """
    np.random.seed(seed)
    idx_v = np.arange(0, n)
    rand_v = list()
    n_rand_i = 0
    n_rand_max = math.factorial(n)-1
    if n_rand_max <= n_rand:
        print("{} random permutations cannot be created. The maximum of n_rand={} is used instead.".format(n_rand, n_rand_max))
        n_rand = n_rand_max
    while n_rand_i < n_rand:
        rand_i = list(np.random.permutation(idx_v))
        if np.all(rand_i == idx_v):
            next
        else:
            if rand_i in rand_v:
                next
            else:
                rand_v.append(rand_i)
                n_rand_i = len(rand_v)
    return rand_v


def permutate_multi_vars(X, rand_index, n_rand, seed=42):
    """
    Function to randomly permutate a multi-dimensional array X `n_rand` times
    at rand_index position.
    """
    x_perm_idx = generate_perms(X.shape[0], n_rand=n_rand, seed=seed)
    X_rand = list()
    for i in np.arange(0, len(x_perm_idx)):
        idx_r = np.array(x_perm_idx[i])
        idx_a = np.tile(np.arange(0, X.shape[0]), (X.shape[1], 1))
        idx_a[rand_index] = idx_r
        idx_a = np.transpose(idx_a)
        X_rand.append(np.take_along_axis(X, indices=idx_a, axis=0))

    return X_rand


def get_res_parallel(y_i, X, s0):
    betas, betas_std, tvals, pvals, tvals_s0, pvals_s0 = regression_workflow(y=y_i, X = X, s0 = s0)
    res = pd.DataFrame()
    res["fc"], res["fc_std"], res["tval"], res["pval"], res["tval_s0"], res["pval_s0"] = [betas, betas_std, tvals, pvals, tvals_s0, pvals_s0]
    return res


def get_perm_res_parallel(y_i, X_rand, s0):
    res_perm_list = regression_workflow_permutation(y_i, X_rand, s0=s0)
    return res_perm_list


def full_regression_analysis(quant_data, annotation, covariates, sample_column='sample_name', n_permutations=4, fdr=0.05, s0=0.05, seed=42):
    data_cols = annotation[sample_column].values
    quant_data = quant_data.dropna().reset_index(drop=True)
    y = quant_data[data_cols].to_numpy().astype('float')
    # @ToDo make sure that columns are sorted correctly!!!
    X = np.array(annotation[covariates])

    # standardize X:
    X_scaled = StandardScaler().fit_transform(X)

    num_cores = multiprocessing.cpu_count()
    res_list = Parallel(n_jobs=num_cores-1)(delayed(get_res_parallel)(y_i=y_i, X=X_scaled, s0=s0) for y_i in y)

    t_limit_dict = {}

    for test_index in np.arange(0, len(covariates)):
        test_covariate = covariates[test_index]
        print(test_covariate)
        X_rand = permutate_multi_vars(X_scaled, rand_index=test_index, n_rand=n_permutations, seed=seed)
        res_perm_list = Parallel(n_jobs=num_cores-1)(delayed(get_perm_res_parallel)(y_i=y_i, X_rand=X_rand, s0=s0) for y_i in y)

        i = test_index + 1
        res_i = [r.iloc[[i]] for r in res_list]
        res_i = pd.concat(res_i)

        # get tvals_s0 list of permutations for covariate i
        res_i_rand = [[0] * len(res_perm_list) for i in range(len(res_perm_list[0]))]
        for a in np.arange(0,len(res_perm_list)):
            for b in np.arange(0, len(res_perm_list[a])):
                #print(res_perm_list[a][b][3])
                res_i_rand[b][a] = res_perm_list[a][b][4][i]

        res_i_rand = [list(np.sort(np.abs(r))) for r in res_i_rand]
        fdr_d_i = get_fdr_stats_across_deltas(res_real=res_i, res_perm=res_i_rand)
        res_i = annotate_fdr_significance(res_real=res_i, stats=fdr_d_i, fdr=fdr)
        t_limit_i = get_tstat_limit(stats=fdr_d_i, fdr=fdr)
        t_limit_dict[test_covariate] = t_limit_i

        # @ToDo: This multiple testing should potentially be done across all covariates together?
        res_i['qval_BH'] = multipletests(res_i.pval, method='fdr_bh')[1]
        res_i['BH_FDR ' + str(int(fdr*100)) + '%'] = ["sig" if abs(x)<=fdr else "non_sig" for x in res_i["qval_BH"]]
        res_i['qval_BH_s0'] = multipletests(res_i.pval_s0, method='fdr_bh')[1]
        res_i['BH_FDR_s0 ' + str(int(fdr*100)) + '%'] = ["sig" if abs(x)<=fdr else "non_sig" for x in res_i["qval_BH_s0"]]

        res_i = res_i.add_prefix(covariates[test_index] + "_")

        quant_data = pd.concat([quant_data.reset_index(drop=True), res_i.reset_index(drop=True)], axis=1)

    return [quant_data, t_limit_dict]


def add_random_covariate(annotation, name='random', n_random=50, seed=42):
    annotation_test = annotation.copy()
    annotation_test.reset_index(drop=True, inplace=True)
    annotation_test[name] = 0
    random.seed(seed)
    rand_true = random.sample(range(0, annotation_test.shape[0]), n_random)
    annotation_test.loc[rand_true, name] = 1
    return annotation_test


def plot_evaluate_seed_and_perm(df, covariates):
    config = {'toImageButtonOptions': {'format': 'svg',
                                      'filename': 'permutation_test',
                                      'scale': 1
                                      }
             }
    for c in covariates:
        fig = px.line(df, x="permutations",
                      y=c,
                      line_dash="seed",
                      line_group="seed",
                      template="simple_white",
                      color_discrete_sequence=['lightgrey'])
        fig.update_traces(mode='lines+markers')
        fig.update_layout(width=620, height=350)
        fig.show(config=config)


def evaluate_seed_and_perm(quant_data,
                           annotation,
                           covariates,
                           perms,
                           seeds,
                           sample_column='sample_name',
                           fdr=0.05,
                           s0=0.05):
    resDF = pd.DataFrame({'seed': np.repeat(seeds, len(perms)),
                          'permutations': np.tile(perms, len(seeds))})
    resDF[covariates] = pd.DataFrame([np.repeat(0, len(covariates))],
                                     index=resDF.index)

    for i in np.arange(0,resDF.shape[0]):
        res, tlim = full_regression_analysis(quant_data=quant_data,
                                             annotation=annotation,
                                             covariates=covariates,
                                             n_permutations=resDF.permutations[i],
                                             sample_column=sample_column,
                                             fdr=fdr,
                                             s0=s0,
                                             seed=resDF.seed[i])

        for c in covariates:
            resDF.loc[i,c] = res[res[c + "_FDR " + str(int(fdr*100)) + "%"] == "sig"].shape[0]
    plot_evaluate_seed_and_perm(resDF, covariates=covariates)
    return resDF


def plot_evaluate_s0s(df, covariates):
    for c in covariates:
        fig = px.line(df, x="s0", y=c, title=c, template="simple_white", color_discrete_sequence=['lightgrey'])
        fig.update_traces(mode='lines+markers')
        fig.update_layout(width=800, height=500)
        fig.show()


def evaluate_s0s(quant_data,
                           annotation,
                           covariates,
                           s0s,
                           sample_column='sample_name',
                           n_permutations=5,
                           seed=42,
                           fdr=0.01):
    resDF = pd.DataFrame({'s0':s0s})
    resDF[covariates] = pd.DataFrame([np.repeat(0, len(covariates))], index=resDF.index)

    for i in np.arange(0,resDF.shape[0]):
        res, tlim = full_regression_analysis(quant_data=quant_data,
                                             annotation=annotation,
                                             covariates=covariates,
                                             sample_column=sample_column,
                                             n_permutations=n_permutations,
                                             fdr=fdr,
                                             s0=resDF.s0[i],
                                             seed=seed)

        for c in covariates:
            resDF.loc[i,c] = res[res[c + "_FDR " + str(int(fdr*100)) + "%"] == "sig"].shape[0]

    plot_evaluate_s0s(resDF, covariates=covariates)

    return resDF


def plot_pval_dist(df, covariates, mode='separate'):
    config = {'toImageButtonOptions': {'format': 'svg',
                                       'filename': 'pvalue_histogram',
                                       'scale': 1
                                      }
             }
    if (mode == 'separate'):
        for c in covariates:
            fig = px.histogram(df,
                               x=c+"_pval",
                               title=c,
                               nbins=50,
                               template='simple_white',
                               color_discrete_sequence=['lightgrey'])
            fig.update_layout(width=400, height=300)
            fig.show(config=config)
    elif (mode == 'joined'):
        df_cov = pd.DataFrame()
        for c in covariates:
            df_c = pd.DataFrame({'p-value': df[c+"_pval"], 'covariate': c})
            df_cov = pd.concat([df_cov, df_c])
        fig = px.histogram(df_cov, x='p-value',
                           color='covariate',
                           barmode='overlay',
                           opacity=0.6,
                           nbins=50,
                           template='simple_white')
        fig.update_layout(width=620, height=350)
        fig.show(config=config)
    else:
        raise ValueError("The mode parameter needs to be either 'separate' or 'joined'.")


def plot_beta_dist(df, covariates):
    for c in covariates:
        fig = px.histogram(df, x=c+"_fc", title=c, template='simple_white', color_discrete_sequence=['lightgrey'])
        fig.update_layout(width=800, height=500)
        fig.show()


def get_replacement_vals(df, threshold, mean_all, sd_all, ds_f):
    """
    Helper function for missing value imputation.
    """
    if df.percent_valid_vals == 100:
        # print(100)
        rep = []
    elif df.percent_valid_vals > threshold:
        # print(70)
        rep = np.random.normal(df.int_mean-(ds_f*df.int_sd), df.int_sd, df.invalid_vals)
    else:
        # print(0)
        rep = np.random.normal(mean_all-(ds_f*sd_all), sd_all, df.invalid_vals)
    return(rep)

def impute_missing_values(data, percent_impute=20, percent_self_impute=70, downshift_factor=1.8):
    """
    Function for missing value imputation. Proteins with less than
    'percent_impute' valid values are removed.
    For proteins with less than 'percent_self_impute' valid values, global mean
    and sd values are used for imputation.
    For proteins with more than 'percent_self_impute' valid values, mean and sd
    values for each specific protein are used.
    The 'downshift_factor' determines by how many standard deviations the mean
    of the imputed distribution is shifted.
    """

    data['int_mean'] = data.filter(regex=("Quantity")).mean(axis=1)
    data['int_sd'] = data.filter(regex=("Quantity")).std(axis=1)
    data['valid_vals'] = data.filter(regex=("Quantity")).count(axis=1)
    data['invalid_vals'] = data.filter(regex=("Quantity")).shape[1]-data.filter(regex=("Quantity")).count(axis=1)
    data['percent_valid_vals'] = 100/data.filter(regex=("Quantity")).shape[1]*data['valid_vals']

    data = data[data.percent_valid_vals >= percent_impute]

    overall_mean = np.mean(data.int_mean)
    overall_sd = np.mean(data.int_sd)

    nan_idx = np.where(data.isna())

    replacement = data.apply(get_replacement_vals, threshold=percent_self_impute, mean_all=overall_mean, sd_all=overall_sd, ds_f = downshift_factor, axis=1)
    replacement = [val for sublist in replacement for val in sublist]

    for i in np.arange(0, len(nan_idx[0])):
        data.iloc[nan_idx[0][i], nan_idx[1][i]] = replacement[i]

    return data



class StatisticUtils:
    def __init__(self) -> None:
        pass
    
    def _add_metadata_column(self, group1_list: list, group2_list: list):

        # create new column in metadata with defined groups
        metadata = self.metadata

        sample_names = metadata[self.sample].to_list()
        misc_samples = list(set(group1_list + group2_list) - set(sample_names))
        if len(misc_samples) > 0:
            raise ValueError(
                f"Sample names: {misc_samples} are not described in Metadata."
            )

        column = "_comparison_column"
        conditons = [
            metadata[self.sample].isin(group1_list),
            metadata[self.sample].isin(group2_list),
        ]
        choices = ["group1", "group2"]
        metadata[column] = np.select(conditons, choices, default=np.nan)
        self.metadata = metadata

        return column, "group1", "group2"



import unittest
import pandas as pd
import numpy as np

from alphastats.loader.MaxQuantLoader import MaxQuantLoader
from alphastats.DataSet import DataSet


class BaseTestDataSet:
    # parent class of test loader for common tests among loaders
    # this is wrapped in a nested class so it doesnt get called separatly when testing
    # plus to avoid multiple inheritance
    class BaseTest(unittest.TestCase):
        def test_go_abundance_corretion(self):
            df = self.obj.go_abundance_correction(
                fg_sample=self.fg_sample, bg_sample=self.bg_sample
            )
            self.assertFalse(df.empty)

        def test_plot_scatter(self):
            df = self.obj.go_abundance_correction(
                fg_sample=self.fg_sample, bg_sample=self.bg_sample
            )
            plot_dict = df.plot_scatter().to_plotly_json()
            # colored in 4 different categories but could change when DB are updated
            self.assertTrue(len(plot_dict.get("data")) > 4)

        def test_plot_bar(self):
            df = self.obj.go_abundance_correction(
                fg_sample=self.fg_sample, bg_sample=self.bg_sample
            )
            plot = df.plot_scatter()

        def test_go_characterize_foreground(self):
            df = self.obj.go_characterize_foreground(
                tax_id=9606, protein_list=self.obj.mat.columns.to_list()[600:700]
            )
            self.assertFalse(df.empty)

        def test_go_compare_samples(self):
            df = self.obj.go_compare_samples(
                fg_sample=self.fg_sample, bg_sample=self.bg_sample
            )
            self.assertTrue(df.empty)

        def test_raise_error_no_evidence(self):
            with self.assertRaises(ValueError):
                self.obj.evidence_df = None
                self.obj.go_abundance_correction(
                    fg_sample=self.fg_sample, bg_sample=self.bg_sample
                )

        def test_go_abundance_correction_with_list(self):
            df = self.obj.go_abundance_correction(
                bg_sample=self.bg_sample, fg_protein_list=self.obj.mat.columns.to_list()[200:300]
            )
            self.assertTrue(df.empty)

        def test_go_genome_list(self):
            df = self.obj.go_genome(protein_list=self.obj.mat.columns.to_list()[600:700])
            self.assertFalse(df.empty)

        def test_go_genome_sample(self):
            df = self.obj.go_genome(fg_sample=self.fg_sample)
            self.assertFalse(df.empty)

        def test_extract_protein_ids(self):
            # test function with different entries
            entry_one = "sp|P0DMV9|HS71B_HUMAN,sp|P0DMV8|HS71A_HUMAN"
            entry_one_protein_id = self.obj._extract_protein_ids(entry=entry_one)
            self.assertEqual(entry_one_protein_id, "P0DMV9;P0DMV8")

            entry_two = "ENSEMBL:ENSBTAP00000007350"
            entry_two_protein_id = self.obj._extract_protein_ids(entry=entry_two)
            self.assertEqual(entry_two_protein_id, "ENSBTAP00000007350")


class TestMaxQuantGODataSet(BaseTestDataSet.BaseTest):
    def setUp(self):
        self.loader = MaxQuantLoader(
            file="testfiles/maxquant_go/proteinGroups.txt",
            evidence_file="testfiles/maxquant_go/evidence.txt",
        )
        evidence_df = pd.read_csv("testfiles/maxquant_go/evidence.txt", sep="\t")
        metadata = pd.DataFrame({"sample": evidence_df["Raw file"].unique().tolist()})
        metadata["experiment"] = np.where(
            metadata["sample"].str.startswith("AC"), "A", "U"
        )

        self.obj = DataSet(
            loader=self.loader, metadata_path=metadata, sample_column="sample",
        )
        self.fg_sample = "AC399"
        self.bg_sample = "UT822"



if __name__ == "__main__":
    unittest.main()



import streamlit as st
import pandas as pd
import io

try:
    from alphastats.gui.utils.ui_helper import sidebar_info
except ModuleNotFoundError:
    from utils.ui_helper import sidebar_info


def display_plotly_figure(plot):
    st.plotly_chart(plot)


def save_plotly(plot, format):
    # Create an in-memory buffer
    buffer = io.BytesIO()
    # Save the figure as a pdf to the buffer
    plot[1].write_image(file=buffer, format=format)
    st.download_button(
        label="Download as " + format, data=buffer, file_name=plot[0] + "." + format
    )


@st.cache_data
def convert_df(df, user_session_id = st.session_state.user_session_id):
    return df.to_csv().encode("utf-8")


def download_preprocessing_info(plot, count):
    preprocesing_dict = plot[1].preprocessing
    df = pd.DataFrame(preprocesing_dict.items())
    filename = "plot" + plot[0] + "preprocessing_info.csv"
    csv = convert_df(df)
    st.download_button(
        "Download DataSet Info as .csv",
        csv,
        filename,
        "text/csv",
        key="preprocessing" + count,
    )


st.markdown("### Results")

sidebar_info()

if "plot_list" in st.session_state:
    for count, plot in enumerate(st.session_state.plot_list):
        count = str(count)

        st.markdown("\n\n")
        st.write(plot[0])

        display_plotly_figure(plot[1])

        col1, col2, col3 = st.columns([1, 1, 1])

        with col1:
            save_plotly(plot, format="pdf")

        with col2:
            save_plotly(plot, format="svg")

        with col3:
            download_preprocessing_info(plot, count)


else:
    st.info("No analysis performed yet.")



import streamlit as st
import pandas as pd

import datetime


def preprocessing():

    c1, c2 = st.columns(2)

    with c1:

        st.markdown(
            "Before analyzing your data, consider normalizing and imputing your data as well as the removal of contaminants. "
            + "A more detailed description about the preprocessing methods can be found in the AlphaPeptStats " 
            + "[documentation](https://alphapeptstats.readthedocs.io/en/main/data_preprocessing.html)."
        )

        with st.form("preprocessing"):
            dataset = st.session_state["dataset"]

            remove_contaminations = st.selectbox(
                f"Remove contaminations annotated in {dataset.filter_columns}",
                options=[True, False],
            )

            subset = st.selectbox(
                "Subset data so it matches with metadata. Remove miscellanous samples in rawinput.",
                options=[True, False],
            )

            remove_samples = st.multiselect(
                "Remove samples from analysis", 
                options=st.session_state.dataset.metadata[st.session_state.dataset.sample].to_list()
            )

            log2_transform = st.selectbox(
                "Log2-transform dataset",
                options=[True, False],
            )

            normalization = st.selectbox(
                "Normalization", options=[None, "zscore", "quantile", "vst", "linear"]
            )

            imputation = st.selectbox(
                "Imputation", options=[None, "mean", "median", "knn", "randomforest"]
            )

            submitted = st.form_submit_button("Submit")

        if submitted:
            if len(remove_samples) == 0:
                remove_samples = None
            
            st.session_state.dataset.preprocess(
                remove_contaminations=remove_contaminations,
                log2_transform=log2_transform,
                remove_samples = remove_samples,
                subset=subset,
                normalization=normalization,
                imputation=imputation,
            )
            preprocessing = st.session_state.dataset.preprocessing_info
            st.info(
                "Data has been processed. "
                + datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
            )
            st.dataframe(
                pd.DataFrame.from_dict(preprocessing, orient="index").astype(str),
                use_container_width=True,
            )
        
        st.markdown("#### Batch correction: correct for technical bias")

        with st.form("Batch correction: correct for technical bias"):
            batch = st.selectbox(
                "Batch", 
                options= st.session_state.dataset.metadata.columns.to_list()
            )
            submit_batch_correction = st.form_submit_button("Submit")
        
        if submit_batch_correction:
            st.session_state.dataset.batch_correction(
                batch=batch
            )
            st.info(
                "Data has been processed. "
                + datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
            )

    
    with c2:

        if submitted:
            st.markdown("**Intensity Distribution after preprocessing per sample**")
            fig_processed = st.session_state.dataset.plot_sampledistribution()
            st.plotly_chart(fig_processed.update_layout(plot_bgcolor="white"), use_container_width=True)
        
        else:
            st.markdown("**Intensity Distribution per sample**")
            fig_none_processed = st.session_state.dataset.plot_sampledistribution()
            st.plotly_chart(fig_none_processed.update_layout(plot_bgcolor="white"), use_container_width=True)
        
    
    reset_steps = st.button("Reset all Preprocessing steps")
        
    if reset_steps:
        reset_preprocessing()




def reset_preprocessing():
    st.session_state.dataset.create_matrix()
    preprocessing = st.session_state.dataset.preprocessing_info
    st.info(
        "Data has been reset. " + datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    )
    st.dataframe(
        pd.DataFrame.from_dict(preprocessing, orient="index").astype(str),
        use_container_width=True,
    )


def main_preprocessing():

    if "dataset" in st.session_state:

        preprocessing()
        
        
    else:
        st.info("Import Data first")


st.markdown("### Preprocessing")


main_preprocessing()


def plot_intensity_distribution():
    st.selectbox(
        "Sample", options=st.session_state.dataset.metadata["sample"].to_list()
    )




import pandas as pd
import scipy
from tqdm import tqdm

class Anova:
    def __init__(self, dataset, column, protein_ids, tukey):
        self.dataset = dataset
        self.column = column
        self.protein_ids = protein_ids
        self.tukey = tukey
        
    def _get_protein_ids_list(self):
        if self.protein_ids == "all":
            self.protein_ids_list = self.dataset.mat.columns.tolist()
        
        elif isinstance(self.protein_ids, str):
            # convert id to list
            self.protein_ids_list = [self.protein_ids]
       
        else:
            self.protein_ids_list = self.protein_ids

    def perform_anova(self) -> pd.DataFrame:
        p_values = self.mat_transpose.apply(
            lambda row: scipy.stats.f_oneway(
                *[row[elem].values.flatten() for elem in self.all_groups]
            )[1],
            axis=1,
        )
        anova_df = pd.DataFrame()
        anova_df[self.dataset.index_column], anova_df["ANOVA_pvalue"] = (
            p_values.index.tolist(),
            p_values.values,
        )
        return anova_df

    def _prepare_data(self):
        #  generated list of list with samples
        subgroup = self.dataset.metadata[self.column].unique().tolist()
        self.all_groups = []
        for sub in subgroup:
            group_list = self.dataset.metadata[self.dataset.metadata[self.column] == sub][
                self.dataset.sample
            ].tolist()
            self.all_groups.append(group_list)

        self.mat_transpose = self.dataset.mat[self.protein_ids_list].transpose()
           
    def _create_tukey_df(self, anova_df: pd.DataFrame) -> pd.DataFrame:
        #  combine tukey results with anova results
        df = (
            self.dataset.mat[self.protein_ids_list]
            .reset_index()
            .rename(columns={"index": self.dataset.sample})
        )
        df = df.merge(self.dataset.metadata, how="inner", on=[self.dataset.sample])
        tukey_df_list = []
        for protein_id in tqdm(self.protein_ids_list):
            tukey_df_list.append(
                self.dataset.tukey_test(df=df, protein_id=protein_id, group=self.column)
            )
        # combine all tukey test results
        tukey_df = pd.concat(tukey_df_list)
        # combine anova and tukey test results
        final_df = anova_df.merge(
            tukey_df[["comparison", "p-tukey", self.dataset.index_column]],
            how="inner",
            on=[self.dataset.index_column],
        )
        final_df = final_df.pivot(
            index=[self.dataset.index_column, "ANOVA_pvalue"],
            columns=["comparison"],
            values="p-tukey",
        )
        return final_df
    
    def perform(self) -> pd.DataFrame:
        self._get_protein_ids_list()
        self._prepare_data()
        anova_df = self.perform_anova()
        
        if self.tukey:
            anova_df = self._create_tukey_df(anova_df=anova_df)
        
        return anova_df




import pandas as pd
import logging
import os
import numpy as np
from alphastats.utils import find_duplicates_in_list
import pkg_resources


class BaseLoader:
    """Parent class of Loaders"""

    def __init__(self, file, intensity_column, index_column, sep):
        """BaseLoader for AlphaPept, MaxQuant, Fragpipe and DIANNLoader

        Args:
            file_path (str): path to file
            sep (str, optional): file separation. Defaults to "\t".
        """

        # self._check_if_file_exists(file=file)
        if isinstance(file, pd.DataFrame):
            self.rawinput = file
        else:
            self.rawinput = pd.read_csv(file, sep=sep, low_memory=False)
        self.intensity_column = intensity_column
        self.index_column = index_column
        self.filter_columns = []
        self.confidence_column = None
        self.software = None
        self.evidence_df = None
        self.gene_names = None
        self.ptm_df = None
        self._add_contamination_column()
        self._check_if_columns_are_present()
        self._read_all_columns_as_string()

    def _check_if_columns_are_present(self):
        """check if given columns present in rawinput"""
        given_columns = list(filter(None, [self.index_column, self.confidence_column]))
        wrong_columns = list(set(given_columns) - set(self.rawinput.columns.to_list()))
        if len(wrong_columns) > 0:
            raise KeyError(
                ", ".join(wrong_columns) + " columns do not exist.\n"
                "Check the documtentation: \n"
                "AlphaPept Format: https://github.com/MannLabs/alphapept \n"
                "DIA-NN Format: https://github.com/vdemichev/DiaNN"
                "FragPipe Format: https://fragpipe.nesvilab.org/docs/tutorial_fragpipe_outputs.html#combined_proteintsv"
                "MaxQuant Format: http://www.coxdocs.org/doku.php?id=maxquant:table:proteingrouptable"
            )
    
    def _read_all_columns_as_string(self):
        self.rawinput.columns = self.rawinput.columns.astype(str)

    def _check_if_indexcolumn_is_unique(self):
        duplicated_values = find_duplicates_in_list(self.rawinput[self.index_column].to_list())
        if len(duplicated_values) > 0:
            # error or warning, duplicates could be resolved with preprocessing/filtering
            logging.warning(
                f"Column {self.index_column} contains duplicated values: "
                + ", ".join(duplicated_values)
            )

    def _check_if_file_exists(self, file):
        if os.path.isfile(file) == False:
            raise OSError(f"{file} does not exist.")

    def _add_contamination_column(self):
        #  load df with potential contamination from fasta file
        contaminations_path = pkg_resources.resource_filename(
            __name__, "../data/contaminations.txt"
        )
        contaminations = pd.read_csv(contaminations_path, sep="\t")
        contaminations_ids = contaminations["Uniprot ID"].to_list()
        #  add column with True False

        self.rawinput["contamination_library"] = np.where(
            self.rawinput[self.index_column].isin(contaminations_ids), True, False
        )
        self.filter_columns = self.filter_columns + ["contamination_library"]

        logging.info(
            "Column 'contamination_library' has been added, to indicate contaminations.\n"
            + "The contaminant library was created by Frankenfield et al."
            + ":https://www.biorxiv.org/content/10.1101/2022.04.27.489766v2.full"
        )



import streamlit as st
st.set_page_config(layout="wide")

try:
    from alphastats.gui.utils.ui_helper import sidebar_info, img_to_bytes
except ModuleNotFoundError:
    from utils.ui_helper import sidebar_info, img_to_bytes
import os
#from PIL import Image

# centering with streamlit is not really centered

from streamlit.runtime import get_instance
from streamlit.runtime.scriptrunner.script_run_context import get_script_run_ctx

runtime = get_instance()
session_id = get_script_run_ctx().session_id
session_info = runtime._session_mgr.get_session_info(session_id)

user_session_id = session_id
st.session_state["user_session_id"] = user_session_id

img_center = """
<head>  
<title> CSS object-position property </title>  
<style> 
img {
  display: block;
  margin-left: auto;
  margin-right: auto;
  width: 65%;
</style>  
</head>  
"""

_this_file = os.path.abspath(__file__)
_this_directory = os.path.dirname(_this_file)
icon = os.path.join(_this_directory, "alphapeptstats_logo.png")

header_html = img_center + "<img src='data:image/png;base64,{}'>".format(
    img_to_bytes(icon)
)

st.markdown(
    header_html,
    unsafe_allow_html=True,
)

##
st.markdown(
    """\n\n
An open-source Python package for the analysis of mass spectrometry based proteomics data 
from the [Mann Group at the University of Copenhagen](https://www.cpr.ku.dk/research/proteomics/mann/).

### How to contribute

If you like this software, you can give us a [star](https://github.com/MannLabs/alphapeptstats/stargazers) to boost 
our visibility! All direct contributions are also welcome. Feel free to post a new [issue](https://github.com/MannLabs/alphapeptstats/issues) 
or clone the repository and create a [pull request](https://github.com/MannLabs/alphapeptstats/pulls) with a new branch. For an even more 
interactive participation, check out the [discussions](https://github.com/MannLabs/alphapeptstats/discussions) and the 
[the Contributors License Agreement](misc/CLA.md).
"""
)

sidebar_info(show_logo=False)

# TODO The emojis on the sidebar menu look ugly but currently there are no other options to add icon in streamlit multipage apps
# https://discuss.streamlit.io/t/icons-for-the-multi-app-page-menu-in-the-sidebar-other-than-emojis/27222
# https://icons.getbootstrap.com/
# https://medium.com/codex/create-a-multi-page-app-with-the-new-streamlit-option-menu-component-3e3edaf7e7ad
#  https://lightrun.com/answers/streamlit-streamlit-set-multipage-app-emoji-in-stpage_config-not-filename



from alphastats.plots.PlotUtils import PlotUtils, plotly_object
from alphastats.utils import ignore_warning, check_for_missing_values

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from functools import lru_cache


class VolcanoPlot(PlotUtils):
    def __init__(
        self, dataset, group1, group2, 
        column=None, method=None, 
        labels=None, min_fc=None, 
        alpha=None, draw_line=None, 
        plot=True, perm=100, fdr=0.05,
        color_list=[]
    ):  
        self.dataset = dataset
        self.group1 = group1
        self.group2 = group2
        self.column = column
        self.method = method
        self.labels = labels
        self.min_fc = min_fc
        self.fdr = fdr
        self.alpha = alpha
        self.draw_line = draw_line
        self.hover_data = None
        self.res = None
        self.pvalue_column = None
        self.perm=perm
        self.color_list = color_list
        self._check_input()
       
        if plot:
            self._perform_differential_expression_analysis()
            self._annotate_result_df()
            self._add_hover_data_columns()
            self._plot()

    def _check_input(self):
        """
        check input and add metadata column if samples are given
        """
        if isinstance(self.group1, list) and isinstance(self.group2, list):
            self.column, self.group1, self.group2 = self.dataset._add_metadata_column(
                self.group1, self.group2
            )

        if self.column is None:
            raise ValueError(
                "Column containing group1 and group2 needs to be specified"
            )

    def _update(self, updated_attributes):
        """
        update attributes using dict
        """
        for key,value in updated_attributes.items():
            setattr(self,key,value)

    @ignore_warning(UserWarning)
    @ignore_warning(RuntimeWarning)
    def _perform_differential_expression_analysis(self):
        """
        wrapper for diff analysis
        """
        if self.method == "wald":
            self._wald()

        elif self.method == "ttest":
            self._ttest()

        elif self.method == "anova":
            self._anova()
        
        elif self.method == "welch-ttest":
            self._welch_ttest()
        
        elif self.method == "paired-ttest":
            self._pairedttest()
        
        elif self.method == "sam":
            self._sam()
        
        # elif self.method == "Multi Covariates":
        #    raise NotImplementedError

        else:
            raise ValueError(
                f"{self.method} is not available."
                + "Please select from 'ttest', 'sam' or 'anova' for anova with follow up tukey or 'wald' for wald-test."
            )

    @lru_cache(maxsize=20)
    def _sam_calculate_fdr_line(self):
        from alphastats.multicova import multicova

        self.fdr_line= multicova.get_fdr_line(
                t_limit=self.tlim_ttest,
                s0=0.05,
                n_x=len(list(self.dataset.metadata[self.dataset.metadata[self.column]==self.group1][self.dataset.sample])),
                n_y=len(list(self.dataset.metadata[self.dataset.metadata[self.column]==self.group2][self.dataset.sample])),
                fc_s = np.arange(0,np.max(np.abs(self.res.log2fc)),np.max(np.abs(self.res.log2fc))/200),
                s_s = np.arange(0.005, 6, 0.0025),
                plot=False
            )
    
    @lru_cache(maxsize=20)
    def _sam(self):
        from alphastats.multicova import multicova

        print(
            "Calculating t-test and permuation based FDR (SAM)... "
        )

        transposed = self.dataset.mat.transpose()

        if self.dataset.preprocessing_info["Normalization"] is None:
             # needs to be lpog2 transformed for fold change calculations
            transposed = transposed.transform(lambda x: np.log2(x))

        transposed[self.dataset.index_column] = transposed.index
        transposed = transposed.reset_index(drop=True)

        res_ttest, tlim_ttest = multicova.perform_ttest_analysis(
            transposed,
            c1 =list(self.dataset.metadata[self.dataset.metadata[self.column]==self.group1][self.dataset.sample]),                                      
            c2 =list(self.dataset.metadata[self.dataset.metadata[self.column]==self.group2][self.dataset.sample]), 
            s0=0.05, 
            n_perm=self.perm,
            fdr=self.fdr,
            id_col=self.dataset.index_column,
            parallelize=True
        )
        
        fdr_column = "FDR"  + str(int(self.fdr*100)) + "%"
        self.res = res_ttest[[self.dataset.index_column, 'fc', 'tval', 'pval', 'tval_s0', 'pval_s0', 'qval']]
        self.res["log2fc"] = res_ttest["fc"]
        self.res["FDR"] = res_ttest[fdr_column]
        self.tlim_ttest = tlim_ttest
        self.pvalue_column = "pval"


    @lru_cache(maxsize=20)
    def _wald(self):

        print(
            "Calculating differential expression analysis using wald test. Fitting generalized linear model..."
        )
        self.res = self.dataset.diff_expression_analysis(
            column=self.column,
            group1=self.group1,
            group2=self.group2,
            method=self.method,
        )
        self.pvalue_column = "qval"

    @lru_cache(maxsize=20)
    def _welch_ttest(self):

        print("Calculating Welchs t-test...")

        self.res = self.dataset.diff_expression_analysis(
            column=self.column,
            group1=self.group1,
            group2=self.group2,
            method=self.method,
        )
        self.pvalue_column = "pval"
    
    @lru_cache(maxsize=20)
    def _ttest(self):

        print("Calculating Students t-test...")

        self.res = self.dataset.diff_expression_analysis(
            column=self.column,
            group1=self.group1,
            group2=self.group2,
            method=self.method,
        )
        self.pvalue_column = "pval"
    
    @lru_cache(maxsize=20)
    def _pairedttest(self):

        print("Calculating paired t-test...")

        self.res = self.dataset.diff_expression_analysis(
            column=self.column,
            group1=self.group1,
            group2=self.group2,
            method=self.method,
        )
        self.pvalue_column = "pval"

    @lru_cache(maxsize=20)
    def _anova(self):

        print("Calculating ANOVA with follow-up tukey test...")

        result_df = self.dataset.anova(
            column=self.column, protein_ids="all", tukey=True
        )

        group1_samples = self.dataset.metadata[
            self.dataset.metadata[self.column] == self.group1
        ][self.dataset.sample].tolist()

        group2_samples = self.dataset.metadata[
            self.dataset.metadata[self.column] == self.group2
        ][self.dataset.sample].tolist()

        mat_transpose = self.dataset.mat.transpose()
        fc = self.dataset._calculate_foldchange(mat_transpose, group1_samples, group2_samples)

        #  check how column is ordered
        self.pvalue_column = self.group1 + " vs. " + self.group2 + " Tukey Test"

        if self.pvalue_column not in result_df.columns:
            self.pvalue_column = self.group2 + " vs. " + self.group1 + " Tukey Test"

        self.res = result_df.reset_index().merge(fc.reset_index(), on=self.dataset.index_column)

    def _add_hover_data_columns(self):
        # additional labeling with gene names
        self.hover_data = [self.dataset.index_column]

        if self.dataset.gene_names is not None:
            self.res = pd.merge(
                self.res,
                self.dataset.rawinput[[self.dataset.gene_names, self.dataset.index_column]],
                on=self.dataset.index_column,
                how="left",
            )
            self.hover_data.append(self.dataset.gene_names)


    def _annotate_result_df(self):
        """
        convert pvalue to log10
        add color labels for up and down regulates
        """
        self.res = self.res[(self.res["log2fc"] < 10) & (self.res["log2fc"] > -10)]
        self.res["-log10(p-value)"] = -np.log10(self.res[self.pvalue_column])
        
        self.alpha = -np.log10(self.alpha)
        # add color variable to plot

        if self.method != "sam":
        
            condition = [
                (self.res["log2fc"] < -self.min_fc) & (self.res["-log10(p-value)"] > self.alpha),
                (self.res["log2fc"] > self.min_fc) & (self.res["-log10(p-value)"] > self.alpha),
            ]

        else:

            condition = [
                (self.res["log2fc"] < 0) & (self.res["FDR"] == "sig"),
                (self.res["log2fc"] > 0) & (self.res["FDR"] == "sig"),
            ]


        value = ["down", "up"]
        self.res["color"] = np.select(condition, value, default="non_sig")   

        if len(self.color_list) > 0:
            self.res["color"] = np.where(self.res[self.dataset.index_column].isin(self.color_list), 
                                          "color", "no_color")   
        

    def _add_labels_plot(self):
        """
        add gene names as hover data if they are given
        """

        if self.dataset.gene_names is not None:
            label_column = self.dataset.gene_names
        else:
            label_column = self.dataset.index_column

        self.res["label"] = np.where(
            self.res.color != "non_sig", self.res[label_column], ""
        )
        #  replace nas with empty string (can cause error when plotting with gene names)
        self.res["label"] = self.res["label"].fillna("")
        self.res = self.res[self.res["label"] != ""]

        for x, y, label_column in self.res[
            ["log2fc", "-log10(p-value)", label_column]
        ].itertuples(index=False):
            self.plot.add_annotation(
                x=x, y=y, text=label_column, showarrow=False, yshift=10
            )

    def _draw_lines_plot(self):
        """
        draw lines using fold change and p-value cut-off, fast method
        """

        self.plot.add_hline(
            y=self.alpha, line_width=1, line_dash="dash", line_color="#8c8c8c"
        )
        self.plot.add_vline(
            x=self.min_fc, line_width=1, line_dash="dash", line_color="#8c8c8c"
        )
        self.plot.add_vline(
            x=-self.min_fc, line_width=1, line_dash="dash", line_color="#8c8c8c"
        )

    def _draw_fdr_line(self):
        """
        Draw fdr line if SAM was applied
        """
        self._sam_calculate_fdr_line()
        
        self.plot.add_trace(go.Scatter(
            x=self.fdr_line[self.fdr_line.fc_s > 0].fc_s,
            y=-np.log10(self.fdr_line[self.fdr_line.fc_s > 0].pvals),
            line_color="black",
            line_shape='spline',
            showlegend=False)
        )
        self.plot.add_trace(go.Scatter(
            x=self.fdr_line[self.fdr_line.fc_s < 0].fc_s,
            y=-np.log10(self.fdr_line[self.fdr_line.fc_s < 0].pvals),
            line_color="black",
            line_shape='spline',
            showlegend=False)
        )
    
    def _color_data_points(self):
         # update coloring
        if len(self.color_list) == 0:
            color_dict = {"non_sig": "#404040", "up": "#B65EAF", "down": "#009599"}
    
        else:
            color_dict = {"no_color": "#404040", "color": "#B65EAF"}
        
        self.plot = self._update_colors_plotly(self.plot, color_dict=color_dict)



    def _plot(self):
        self.plot = px.scatter(
            self.res,
            x="log2fc",
            y="-log10(p-value)",
            color="color",
            hover_data=self.hover_data,
        )
        
        # update coloring
        self._color_data_points()

        if self.labels:
            self._add_labels_plot()
        
        if self.draw_line:
            if self.method == "sam":
                self._draw_fdr_line()
            else:
                self._draw_lines_plot()

        self.plot.update_layout(showlegend=False)
        self.plot.update_layout(width=600, height=700)

        #  save plotting data in figure object
        self.plot = plotly_object(self.plot)
        self.plot = self._update_figure_attributes(
            figure_object=self.plot, 
            plotting_data=self.res, 
            preprocessing_info=self.dataset.preprocessing_info, 
            method=self.method
        )
    

   



from random import sample
import pandas as pd
import numpy as np
import logging
import warnings
import plotly

from alphastats.loader.AlphaPeptLoader import AlphaPeptLoader
from alphastats.loader.DIANNLoader import DIANNLoader
from alphastats.loader.FragPipeLoader import FragPipeLoader
from alphastats.loader.MaxQuantLoader import MaxQuantLoader
from alphastats.loader.SpectronautLoader import SpectronautLoader
from alphastats.loader.mzTabLoader import mzTabLoader

from alphastats.DataSet_Plot import Plot
from alphastats.DataSet_Preprocess import Preprocess
from alphastats.DataSet_Pathway import Enrichment
from alphastats.DataSet_Statistics import Statistics
from alphastats.utils import LoaderError

# remove warning from openpyxl
# only appears on mac
warnings.filterwarnings("ignore", category=UserWarning, module="openpyxl")

plotly.io.templates["alphastats_colors"] = plotly.graph_objects.layout.Template(
    layout=plotly.graph_objects.Layout(
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        colorway=[
            "#009599",
            "#005358",
            "#772173",
            "#B65EAF",  # pink
            "#A73A00",
            "#6490C1",
            "#FF894F",
            "#2B5E8B",
            "#A87F32",
        ],
    )
)

plotly.io.templates.default = "simple_white+alphastats_colors"


class DataSet(Preprocess, Statistics, Plot, Enrichment):
    """Analysis Object"""

    def __init__(self, loader, metadata_path=None, sample_column=None):
        """Create DataSet

        Args:
            loader (_type_): loader of class AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader, SpectronautLoader
            metadata_path (str, optional): path to metadata file. Defaults to None.
            sample_column (str, optional): column in metadata file indicating the sample IDs. Defaults to None.

        """
        self._check_loader(loader=loader)
        #  load data from loader object
        self.loader = loader
        self.rawinput = loader.rawinput
        self.software = loader.software
        self.index_column = loader.index_column
        self.intensity_column = loader.intensity_column
        self.filter_columns = loader.filter_columns
        self.evidence_df = loader.evidence_df
        self.gene_names = loader.gene_names

        # include filtering before
        self.create_matrix()
        self._check_matrix_values()
        self.metadata = None
        if metadata_path is not None:
            self.sample = sample_column
            self.load_metadata(file_path=metadata_path)
            self._remove_misc_samples_in_metadata()

        else:
            self._create_metadata()

        # save preprocessing settings
        self.preprocessing_info = self._save_dataset_info()
        self.preprocessed = False

        print("DataSet has been created.")
        self.overview()

    def _create_metadata(self):
        samples = list(self.mat.index)
        self.metadata = pd.DataFrame({"sample": samples})
        self.sample = "sample"

    def _check_loader(self, loader):
        """Checks if the Loader is from class AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader

        Args:
            loader : loader
        """
        if not isinstance(
            loader, (AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader, SpectronautLoader, mzTabLoader)
        ):
            raise LoaderError(
                "loader must be from class: AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader or SpectronautLoader"
            )

        if not isinstance(loader.rawinput, pd.DataFrame) or loader.rawinput.empty:
            raise ValueError(
                "Error in rawinput, consider reloading your data with: AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader, SpectronautLoader"
            )

        if not isinstance(loader.index_column, str):
            raise ValueError(
                "Invalid index_column: consider reloading your data with: AlphaPeptLoader, MaxQuantLoader, DIANNLoader, FragPipeLoader, SpectronautLoader"
            )

    def _check_matrix_values(self):
        if np.isinf(self.mat).values.sum() > 0:
            logging.warning("Data contains infinite values.")

    def _remove_misc_samples_in_metadata(self):
        samples_matrix = self.mat.index.to_list()
        samples_metadata = self.metadata[self.sample].to_list()
        misc_samples = list(set(samples_metadata) - set(samples_matrix))
        if len(misc_samples) > 0:
            self.metadata = self.metadata[
                ~self.metadata[self.sample].isin(misc_samples)
            ]
            logging.warning(
                f"{misc_samples} are not described in the protein data and"
                "are removed from the metadata."
            )

    def create_matrix(self):
        """
        Creates a matrix of the Outputfile, with columns displaying features (Proteins) and
        rows the samples.
        """

        regex_find_intensity_columns = self.intensity_column.replace("[sample]", ".*")
        
        df = self.rawinput
        df = df.set_index(self.index_column)
        df = df.filter(regex=(regex_find_intensity_columns), axis=1)
        # remove Intensity so only sample names remain
        substring_to_remove = regex_find_intensity_columns.replace(".*", "")
        df.columns = df.columns.str.replace(substring_to_remove, "")
        # transpose dataframe
        mat = df.transpose()
        # remove proteins with only zero
        self.mat = mat.loc[:, (mat != 0).any(axis=0)]
        # reset preproccessing info
        self.preprocessing_info = self._save_dataset_info()
        self.preprocessed = False
        self.rawmat = mat

    def load_metadata(self, file_path):
        """Load metadata either xlsx, txt, csv or txt file

        Args:
            file_path (str): path to metadata file
        """
        if isinstance(file_path, pd.DataFrame):
            df = file_path
        #  loading file needs to be more beautiful
        elif file_path.endswith(".xlsx"):
            df = pd.read_excel(file_path)
            # find robust way to detect file format
            # else give file separation as variable
        elif file_path.endswith(".txt") or file_path.endswith(".tsv"):
            df = pd.read_csv(file_path, delimiter="\t")
        elif file_path.endswith(".csv"):
            df = pd.read_csv(file_path)
        else:
            df = None
            logging.warning(
                "WARNING: Metadata could not be read. \nMetadata has to be a .xslx, .tsv, .csv or .txt file"
            )
            return
        if df is not None and self.sample not in df.columns:
            logging.error(f"sample_column: {self.sample} not found in {file_path}")

        # check whether sample labeling matches protein data
        #  warnings.warn("WARNING: Sample names do not match sample labelling in protein data")
        df.columns = df.columns.astype(str)
        self.metadata = df

    def _save_dataset_info(self):
        n_proteingroups = self.mat.shape[1]
        preprocessing_dict = {
            "Raw data number of Protein Groups": n_proteingroups,
            "Matrix: Number of ProteinIDs/ProteinGroups": self.mat.shape[1],
            "Matrix: Number of samples": self.mat.shape[0],
            "Intensity used for analysis": self.intensity_column,
            "Log2-transformed": False,
            "Normalization": None,
            "Imputation": None,
            "Contaminations have been removed": False,
            "Contamination columns": self.filter_columns,
            "Number of removed ProteinGroups due to contaminaton": 0,
        }
        return preprocessing_dict

    def overview(self):
        """Print overview of the DataSet"""
        dataset_overview = (
            "Attributes of the DataSet can be accessed using: \n"
            + "DataSet.rawinput:\t Raw Protein data.\n"
            + "DataSet.mat:\tProcessed data matrix with ProteinIDs/ProteinGroups as columns and samples as rows. All computations are performed on this matrix.\n"
            + "DataSet.metadata:\tMetadata for the samples in the matrix. Metadata will be matched with DataSet.mat when needed (for instance Volcano Plot)."
        )
        print(dataset_overview)



import plotly
import seaborn as sns
import plotly.graph_objects as go

class PlotUtils:
    def __init__(self) -> None:
        pass

    @staticmethod
    def _update_colors_plotly(fig, color_dict):
        # plotly doesnt allow to assign color to certain group
        # update instead the figure in form of a dict
        # color_dict with group_variable/legendgroup as key, and corresponding color as value
        fig_dict = fig.to_plotly_json()
        data_dict_list = fig_dict.get("data")
        for count, group in enumerate(data_dict_list):
            group_variable = group.get("legendgroup")
            group_color = color_dict.get(group_variable)
            fig_dict["data"][count]["marker"]["color"] = group_color
        # convert dict back to plotly figure
        return go.Figure(fig_dict)

    def _update_figure_attributes(self, figure_object, plotting_data, preprocessing_info, method=None):
        setattr(figure_object, "plotting_data", plotting_data)
        setattr(figure_object, "preprocessing", preprocessing_info)
        setattr(figure_object, "method", method)
        return figure_object


class plotly_object(plotly.graph_objs._figure.Figure):
    plotting_data = None
    preprocessing = None
    method = None


class seaborn_object(sns.matrix.ClusterGrid):
    plotting_data = None
    preprocessing = None
    method = None



import streamlit as st

import pandas as pd
import io


try:
    from alphastats.gui.utils.ui_helper import sidebar_info
    from alphastats.gui.utils.analysis_helper import (
    get_analysis,
)  

except ModuleNotFoundError:
    from utils.ui_helper import sidebar_info
    from utils.analysis_helper import (
    get_analysis,
)  

def check_if_options_are_loaded(f):
    """
    decorator to check whether analysis options are loaded
    """

    def inner(*args, **kwargs):

        if hasattr(st.session_state, "plotting_options") is False:
            alphastats.gui.utils.analysis_helper.load_options()

        return f(*args, **kwargs)

    return inner


def display_figure(plot):
    """
    display plotly or seaborn figure
    """
    try:
        st.plotly_chart(plot.update_layout(plot_bgcolor="white"))
    except:
        st.pyplot(plot)


def save_plot_to_session_state(plot, method):
    """
    save plot with method to session state to retrieve old results
    """
    st.session_state["plot_list"] += [(method, plot)]


def display_df(df):
    mask = df.applymap(type) != bool
    d = {True: "TRUE", False: "FALSE"}
    df = df.where(mask, df.replace(d))
    st.dataframe(df)


def download_figure(obj, format, plotting_library="plotly"):
    """
    download plotly figure
    """

    plot = obj[1]
    filename = obj[0] + "." + format

    buffer = io.BytesIO()

    if plotting_library == "plotly":
        # Save the figure as a pdf to the buffer
        plot.write_image(file=buffer, format=format)

    else:
        plot.savefig(buffer, format=format)

    st.download_button(label="Download as " + format, data=buffer, file_name=filename)


@st.cache_data
def convert_df(df, user_session_id = st.session_state.user_session_id):
    return df.to_csv().encode("utf-8")


def download_preprocessing_info(plot):
    preprocesing_dict = plot[1].preprocessing
    df = pd.DataFrame(preprocesing_dict.items())
    filename = "plot" + plot[0] + "preprocessing_info.csv"
    csv = convert_df(df)
    st.download_button(
        "Download DataSet Info as .csv",
        csv,
        filename,
        "text/csv",
        key="preprocessing",
    )


@check_if_options_are_loaded
def select_analysis():
    """
    select box
    loads keys from option dicts
    """
    method = st.selectbox(
        "Analysis",
        options=list(st.session_state.plotting_options.keys())
        + list(st.session_state.statistic_options.keys()),
    )
    return method


st.markdown("### Analysis")

sidebar_info()


# set background to white so downloaded pngs dont have grey background
styl = f"""
    <style>
        .css-jc5rf5 {{
            position: absolute;
            background: rgb(255, 255, 255);
            color: rgb(48, 46, 48);
            inset: 0px;
            overflow: hidden;
        }}
    </style>
    """
st.markdown(styl, unsafe_allow_html=True)


if "plot_list" not in st.session_state:
    st.session_state["plot_list"] = []


if "dataset" in st.session_state:

    c1, c2 = st.columns((1,2))

    plot_to_display = False
    df_to_display = False

    with c1:

        method = select_analysis()

        if method in st.session_state.plotting_options.keys():

            analysis_result = get_analysis(
                method=method, options_dict=st.session_state.plotting_options
            )
            plot_to_display = True
        
        elif method in st.session_state.statistic_options.keys():

            analysis_result = get_analysis(
                method=method, options_dict=st.session_state.statistic_options
            )
            df_to_display = True

        st.markdown("")
        st.markdown("")
        st.markdown("")
        st.markdown("")
        

    with c2:

        # --- Plot -------------------------------------------------------

        if analysis_result is not None and method != "Clustermap" and plot_to_display:
            
            display_figure(analysis_result)

            save_plot_to_session_state(analysis_result, method)

            method_plot = [method, analysis_result]

            
        elif method == "Clustermap":

            st.write("Download Figure to see full size.")

            display_figure(analysis_result)

            save_plot_to_session_state(analysis_result, method)


        # --- STATISTICAL ANALYSIS -------------------------------------------------------

        elif analysis_result is not None and df_to_display:

            display_df(analysis_result)

            filename = method + ".csv"
            csv = convert_df(analysis_result)

            
    
    if analysis_result is not None and method != "Clustermap" and plot_to_display:
        col1, col2, col3 = st.columns([1, 1, 1])

        with col1:
            download_figure(method_plot, format="pdf")

        with col2:
            download_figure(method_plot, format="svg")

        with col3:
            download_preprocessing_info(method_plot)

    
    elif analysis_result is not None and df_to_display:
        col1, col2, col3 = st.columns([1, 1, 1])

        with col1:
            download_figure(method_plot, format="pdf", plotting_library="seaborn")

        with col2:
            download_figure(method_plot, format="svg", plotting_library="seaborn")

        with col3:
             download_preprocessing_info(method_plot)
    
    elif analysis_result is not None and df_to_display:
        st.download_button(
                "Download as .csv", csv, filename, "text/csv", key="download-csv"
        )
            



else:
    st.info("Import Data first")



from random import random
import pandas as pd
import sklearn
import logging
import numpy as np
import sklearn.ensemble
import sklearn.impute
from alphastats.utils import ignore_warning
from sklearn.experimental import enable_iterative_imputer
import itertools



class Preprocess:
    def _remove_sampels(self, sample_list):
        # exclude samples for analysis
        self.mat = self.mat.drop(sample_list)
        self.metadata = self.metadata[~self.metadata[self.sample].isin(sample_list)]

    def _subset(self):
        # filter matrix so only samples that are described in metadata
        # also found in matrix
        self.preprocessing_info.update(
            {"Matrix: Number of samples": self.metadata.shape[0]}
        )
        return self.mat[self.mat.index.isin(self.metadata[self.sample].tolist())]

    def preprocess_print_info(self):
        """Print summary of preprocessing steps"""
        print(pd.DataFrame(self.preprocessing_info.items()))

    def _filter(self):
        if len(self.filter_columns) == 0:
            logging.info("No columns to filter.")
            return

        if self.preprocessing_info.get("Contaminations have been removed") == True:
            logging.info("Contaminatons have already been filtered.")
            return

        #  print column names with contamination
        protein_groups_to_remove = self.rawinput[
            (self.rawinput[self.filter_columns] == True).any(axis=1)
        ][self.index_column].tolist()

        protein_groups_to_remove = list(
            set(protein_groups_to_remove) & set(self.mat.columns.to_list())
        )

        # remove columns with protein groups
        self.mat = self.mat.drop(protein_groups_to_remove, axis=1)

        self.preprocessing_info.update(
            {
                "Number of removed ProteinGroups due to contaminaton": len(
                    protein_groups_to_remove
                ),
                "Contaminations have been removed": True,
                "Matrix: Number of ProteinIDs/ProteinGroups": self.mat.shape[1],
            }
        )

        filter_info = (
            f"Contaminations indicated in following columns: {self.filter_columns} were removed. "
            f"In total {str(len(protein_groups_to_remove))} observations have been removed."
        )
        logging.info(filter_info)

    @ignore_warning(RuntimeWarning)
    @ignore_warning(UserWarning)
    def _imputation(self, method):
        # remove ProteinGroups with only NA before
        protein_group_na = self.mat.columns[self.mat.isna().all()].tolist()

        if len(protein_group_na) > 0:
            self.mat = self.mat.drop(protein_group_na, axis=1)
            logging.info(
                f" {len(protein_group_na)} Protein Groups were removed due to missing values."
            )

        logging.info("Imputing data...")

        if method == "mean":
            imp = sklearn.impute.SimpleImputer(missing_values=np.nan, strategy="mean")
            imputation_array = imp.fit_transform(self.mat.values)

        elif method == "median":
            imp = sklearn.impute.SimpleImputer(missing_values=np.nan, strategy="median")
            imputation_array = imp.fit_transform(self.mat.values)

        elif method == "knn":
            # change for text
            method = "k-Nearest Neighbor"
            imp = sklearn.impute.KNNImputer(n_neighbors=2, weights="uniform")
            imputation_array = imp.fit_transform(self.mat.values)

        elif method == "randomforest":
            randomforest = sklearn.ensemble.RandomForestRegressor(
                max_depth=10,
                bootstrap=True,
                max_samples=0.5,
                n_jobs=2,
                random_state=0,
                verbose=0,  #  random forest takes a while print progress
            )
            imp = sklearn.impute.IterativeImputer(
                random_state=0, estimator=randomforest
            )

            # the random forest imputer doesnt work with float32/float16..
            #  so the values are multiplied and converted to integers
            array_multi_mio = self.mat.values * 1000000
            array_int = array_multi_mio.astype("int")

            imputation_array = imp.fit_transform(array_int)
            imputation_array = imputation_array / 1000000

        else:
            raise ValueError(
                "Imputation method: {method} is invalid."
                "Choose 'mean'. 'median', 'knn' (for k-nearest Neighbors) or "
                "'randomforest' for random forest imputation."
            )

        self.mat = pd.DataFrame(
            imputation_array, index=self.mat.index, columns=self.mat.columns
        )
        self.preprocessing_info.update({"Imputation": method})

    @ignore_warning(UserWarning)
    @ignore_warning(RuntimeWarning)
    def _normalization(self, method):
    
        if method == "zscore":
            scaler = sklearn.preprocessing.StandardScaler()
            normalized_array = scaler.fit_transform(self.mat.values)

        elif method == "quantile":
            qt = sklearn.preprocessing.QuantileTransformer(random_state=0)
            normalized_array = qt.fit_transform(self.mat.values)

        elif method == "linear":
            normalized_array = sklearn.preprocessing.normalize(
                self.mat.values, norm="l2"
            )

        elif method == "vst":
            scaler = sklearn.preprocessing.PowerTransformer()
            normalized_array = scaler.fit_transform(self.mat.values)

        else:
            raise ValueError(
                "Normalization method: {method} is invalid"
                "Choose from 'zscore', 'quantile', 'linear' normalization. or 'vst' for variance stabilization transformation"
            )

        self.mat = pd.DataFrame(
            normalized_array, index=self.mat.index, columns=self.mat.columns
        )

        self.preprocessing_info.update({"Normalization": method})

    def reset_preprocessing(self):
        """ Reset all preprocessing steps
        """
        # reset all preprocessing steps
        self.create_matrix()
        print("All preprocessing steps are reset.")
    
    def _compare_preprocessing_modes(self, func, params_for_func) -> list:
        dataset = self
        imputation_methods = ["mean", "median", "knn"]
        normalization_methods = ["zscore", "quantile", "vst"]
        preprocessing_modes = list(itertools.product(normalization_methods, imputation_methods))

        results_list = []

        del params_for_func["compare_preprocessing_modes"]
        params_for_func["dataset"] = params_for_func.pop("self")

        for preprocessing_mode in preprocessing_modes:
            # reset preprocessing
            dataset.reset_preprocessing()
            print(f"Normalization {preprocessing_mode[0]}, Imputation {str(preprocessing_mode[1])}")
            
            dataset.preprocess(
                subset=True,
                normalization = preprocessing_mode[0],
                imputation = preprocessing_mode[1]
            )
            
            res = func(**params_for_func)
            results_list.append(res)
        
        return results_list

    def _log2_transform(self):
        self.mat = np.log2(self.mat + 0.1)
        self.preprocessing_info.update({"Log2-transformed": True})
        print("Data has been log2-transformed.")
    
    def batch_correction(self, batch:str):
        """Correct for technical bias/batch effects
        Behdenna A, Haziza J, Azencot CA and Nordor A. (2020) pyComBat, a Python tool for batch effects correction in high-throughput molecular data using empirical Bayes methods. bioRxiv doi: 10.1101/2020.03.17.995431

        Args:
            batch (str): column name in the metadata describing the different batches
        """
        import combat
        from combat.pycombat import pycombat
        data = self.mat.transpose()
        series_of_batches = self.metadata.set_index(self.sample).reindex(data.columns.to_list())[batch]
        self.mat = pycombat(data=data, batch=series_of_batches).transpose()

    @ignore_warning(RuntimeWarning)
    def preprocess(
        self,
        log2_transform: bool=True,
        remove_contaminations: bool=False,
        subset: bool=False,
        normalization: str=None,
        imputation: str=None,
        remove_samples: list=None,
    ):
        """Preprocess Protein data

        Removal of contaminations:

        Removes all observations, that were identified as contaminations.

        Normalization:

        "zscore", "quantile", "linear", "vst"

        Normalize data using either zscore, quantile or linear (using l2 norm) Normalization.

        Z-score normalization equals standaridzation using StandardScaler:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html

        Variance stabilization transformation uses:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PowerTransformer.html

        For more information visit.
        Sklearn: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.normalize.html

        Imputation:

        "mean", "median", "knn" or "randomforest"
        For more information visit:

        SimpleImputer: https://scikit-learn.org/stable/modules/generated/sklearn.impute.SimpleImputer.html

        k-Nearest Neighbors Imputation: https://scikit-learn.org/stable/modules/impute.html#impute

        Random Forest Imputation: https://scikit-learn.org/stable/auto_examples/impute/plot_iterative_imputer_variants_comparison.html
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html#sklearn.ensemble.RandomForestRegressor

        Args:
            remove_contaminations (bool, optional): remove ProteinGroups that are identified as contamination.
            log2_transform (bool, optional): Log2 transform data. Default to True.
            normalization (str, optional): method to normalize data: either "zscore", "quantile", "linear". Defaults to None.
            remove_samples (list, optional): list with sample ids to remove. Defaults to None.
            imputation (str, optional):  method to impute data: either "mean", "median", "knn" or "randomforest". Defaults to None.
            subset (bool, optional): filter matrix so only samples that are described in metadata found in matrix. Defaults to False.
        """
        if remove_contaminations:
            self._filter()
        
        if remove_samples is not None:
            self._remove_sampels(sample_list=remove_samples)

        if subset:
            self.mat = self._subset()
        
        if log2_transform:
            self._log2_transform()

        if normalization is not None:
            self._normalization(method=normalization)

        if imputation is not None:
            self._imputation(method=imputation)

        self.mat = self.mat.loc[:, (self.mat != 0).any(axis=0)]
        self.preprocessed = True
